searchData={"items":[{"type":"module","title":"Phoenix.Sync","doc":"Real-time sync for Postgres-backed Phoenix applications.\n\nSee the [docs](../../README.md) for more information.","ref":"Phoenix.Sync.html"},{"type":"function","title":"Phoenix.Sync.client!/0","doc":"","ref":"Phoenix.Sync.html#client!/0"},{"type":"type","title":"Phoenix.Sync.param_override/0","doc":"","ref":"Phoenix.Sync.html#t:param_override/0"},{"type":"type","title":"Phoenix.Sync.param_overrides/0","doc":"","ref":"Phoenix.Sync.html#t:param_overrides/0"},{"type":"function","title":"Phoenix.Sync.plug_opts/0","doc":"Returns the required adapter configuration for your Phoenix Endpoint or\n`Plug.Router`.","ref":"Phoenix.Sync.html#plug_opts/0"},{"type":"function","title":"Phoenix - Phoenix.Sync.plug_opts/0","doc":"Configure your endpoint with the configuration at runtime by passing the\n`phoenix_sync` configuration to your endpoint in the `Application.start/2`\ncallback:\n\n    def start(_type, _args) do\n      children = [\n        # ...\n        {MyAppWeb.Endpoint, phoenix_sync: Phoenix.Sync.plug_opts()}\n      ]\n    end","ref":"Phoenix.Sync.html#plug_opts/0-phoenix"},{"type":"function","title":"Plug - Phoenix.Sync.plug_opts/0","doc":"Add the configuration to the Plug opts in your server configuration:\n\n    children = [\n      {Bandit, plug: {MyApp.Router, phoenix_sync: Phoenix.Sync.plug_opts()}}\n    ]\n\nYour `Plug.Router` must be configured with\n[`copy_opts_to_assign`](https://hexdocs.pm/plug/Plug.Builder.html#module-options) and you should `use` the rele\n\n    defmodule MyApp.Router do\n      use Plug.Router, copy_opts_to_assign: :options\n\n      use Phoenix.Sync.Controller\n      use Phoenix.Sync.Router\n\n      plug :match\n      plug :dispatch\n\n      sync \"/shapes/todos\", Todos.Todo\n\n      get \"/shapes/user-todos\" do\n        %{\"user_id\" => user_id} = conn.params\n        sync_render(conn, from(t in Todos.Todo, where: t.owner_id == ^user_id)\n      end\n    end","ref":"Phoenix.Sync.html#plug_opts/0-plug"},{"type":"type","title":"Phoenix.Sync.shape_definition/0","doc":"","ref":"Phoenix.Sync.html#t:shape_definition/0"},{"type":"type","title":"Phoenix.Sync.shape_specification/0","doc":"","ref":"Phoenix.Sync.html#t:shape_specification/0"},{"type":"module","title":"Phoenix.Sync.Client","doc":"Low level Elixir client. Converts an `Ecto.Query` into an Elixir `Stream`:\n\n```elixir\nstream = Phoenix.Sync.Client.stream(Todos.Todo)\n\nstream =\n  Ecto.Query.from(t in Todos.Todo, where: t.completed == false)\n  |> Phoenix.Sync.Client.stream()\n```","ref":"Phoenix.Sync.Client.html"},{"type":"function","title":"Phoenix.Sync.Client.new/0","doc":"Create a new sync client based on the `:phoenix_sync` configuration.","ref":"Phoenix.Sync.Client.html#new/0"},{"type":"function","title":"Phoenix.Sync.Client.new/1","doc":"Create a sync client using the given options.\n\nIf the integration mode is set to `:embedded` and Electric is installed\nthen this will configure the client to retrieve data using the internal\nElixir APIs.\n\nFor the `:http` mode, then you must also configure a URL specifying an\nElectric API server:\n\n    config :phoenix_sync,\n      mode: :http,\n      url: \"https://api.electric-sql.cloud\"\n\nThis client can then generate streams for use in your Elixir applications:\n\n    {:ok, client} = Phoenix.Sync.Client.new()\n    stream = Electric.Client.stream(client, Todos.Todo)\n    for msg <- stream, do: IO.inspect(msg)\n\nAlternatively use `stream/1` which wraps this functionality.","ref":"Phoenix.Sync.Client.html#new/1"},{"type":"function","title":"Phoenix.Sync.Client.new!/0","doc":"Create a new sync client based on the application configuration or raise if\nthe config is invalid.\n\n    client = Phoenix.Sync.Client.new!()\n\nSee `new/0`.","ref":"Phoenix.Sync.Client.html#new!/0"},{"type":"function","title":"Phoenix.Sync.Client.new!/1","doc":"Create a new sync client based on the given opts or raise if\nthe config is invalid.\n\n    client = Phoenix.Sync.Client.new!(mode: :embedded)\n\nSee `new/1`.","ref":"Phoenix.Sync.Client.html#new!/1"},{"type":"function","title":"Phoenix.Sync.Client.stream/2","doc":"Return a sync stream for the given shape.","ref":"Phoenix.Sync.Client.html#stream/2"},{"type":"function","title":"Examples - Phoenix.Sync.Client.stream/2","doc":"# stream updates for the Todo schema\n    stream = Phoenix.Sync.Client.stream(MyApp.Todos.Todo)\n\n    # stream the results of an ecto query\n    stream = Phoenix.Sync.Client.stream(from(t in MyApp.Todos.Todo, where: t.completed == true))\n\n    # create a stream based on a shape definition\n    stream = Phoenix.Sync.Client.stream(\n      table: \"todos\",\n      where: \"completed = false\",\n      columns: [\"id\", \"title\"]\n    )\n\n    # once you have a stream, consume it as usual\n    Enum.each(stream, &IO.inspect/1)","ref":"Phoenix.Sync.Client.html#stream/2-examples"},{"type":"function","title":"Ecto vs keyword shapes - Phoenix.Sync.Client.stream/2","doc":"Streams defined using an Ecto query or schema will return data wrapped in\nthe appropriate schema struct, with values cast to the appropriate\nElixir/Ecto types, rather than raw column data in the form `%{\"column_name\"\n=> \"column_value\"}`.","ref":"Phoenix.Sync.Client.html#stream/2-ecto-vs-keyword-shapes"},{"type":"module","title":"Phoenix.Sync.Controller","doc":"Provides controller-level integration with sync streams.\n\nUnlike `Phoenix.Sync.Router.sync/2`, which only permits static shape\ndefinitions, in a controller you can use request and session information to\nfilter your data.","ref":"Phoenix.Sync.Controller.html"},{"type":"module","title":"Phoenix Example - Phoenix.Sync.Controller","doc":"defmodule MyAppWeb.TodoController do\n      use Phoenix.Controller, formats: [:html, :json]\n\n      import Elixir.Phoenix.Sync.Controller\n\n      alias MyApp.Todos\n\n      def all(conn, %{\"user_id\" => user_id} = params) do\n        sync_render(\n          conn,\n          params,\n          from(t in Todos.Todo, where: t.owner_id == ^user_id)\n        )\n      end\n    end","ref":"Phoenix.Sync.Controller.html#module-phoenix-example"},{"type":"module","title":"Plug Example - Phoenix.Sync.Controller","doc":"You should `use Elixir.Phoenix.Sync.Controller` in your `Plug.Router`, then within your route\nyou can use the `sync_render/2` function.\n\n    defmodule MyPlugApp.Router do\n      use Plug.Router, copy_opts_to_assign: :options\n      use Elixir.Phoenix.Sync.Controller\n\n      plug :match\n      plug :dispatch\n\n      get \"/todos\" do\n        sync_render(conn, MyPlugApp.Todos.Todo)\n      end\n    end","ref":"Phoenix.Sync.Controller.html#module-plug-example"},{"type":"function","title":"Phoenix.Sync.Controller.sync_render/3","doc":"Return the sync events for the given shape as a `Plug.Conn` response.","ref":"Phoenix.Sync.Controller.html#sync_render/3"},{"type":"module","title":"Phoenix.Sync.Electric","doc":"A `Plug.Router` and `Phoenix.Router` compatible Plug handler that allows\nyou to mount the full Electric shape api into your application.\n\nUnlike `Phoenix.Sync.Router.sync/2` this allows your app to serve\nshapes defined by `table` parameters, much like the Electric application.\n\nThe advantage is that you're free to put your own authentication and\nauthorization Plugs in front of this endpoint, integrating the auth for\nyour shapes API with the rest of your app.","ref":"Phoenix.Sync.Electric.html"},{"type":"module","title":"Configuration - Phoenix.Sync.Electric","doc":"Before configuring your router, you must install and configure the\n`:phoenix_sync` application.\n\nSee the documentation for [embedding electric](readme.html#installation-and-configuration) for\ndetails on embedding Electric into your Elixir application.","ref":"Phoenix.Sync.Electric.html#module-configuration"},{"type":"module","title":"Plug Integration - Phoenix.Sync.Electric","doc":"Mount this Plug into your router using `Plug.Router.forward/2`.\n\n    defmodule MyRouter do\n      use Plug.Router, copy_opts_to_assign: :config\n      use Phoenix.Sync.Electric\n\n      plug :match\n      plug :dispatch\n\n      forward \"/shapes\",\n        to: Phoenix.Sync.Electric,\n        init_opts: [opts_in_assign: :config]\n    end\n\nYou **must** configure your `Plug.Router` with `copy_opts_to_assign` and\npass the key you configure here (in this case `:config`) to the\n`Phoenix.Sync.Electric` plug in it's `init_opts`.\n\nIn your application, build your Electric confguration using\n`Phoenix.Sync.plug_opts()` and pass the result to your router as\n`phoenix_sync`:\n\n    # in application.ex\n    def start(_type, _args) do\n      children = [\n        {Bandit, plug: {MyRouter, phoenix_sync: Phoenix.Sync.plug_opts()}, port: 4000}\n      ]\n\n      Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor)\n    end","ref":"Phoenix.Sync.Electric.html#module-plug-integration"},{"type":"module","title":"Phoenix Integration - Phoenix.Sync.Electric","doc":"Use `Phoenix.Router.forward/2` in your router:\n\n    defmodule MyAppWeb.Router do\n      use Phoenix.Router\n\n      pipeline :shapes do\n        # your authz plugs\n      end\n\n      scope \"/shapes\" do\n        pipe_through [:shapes]\n\n        forward \"/\", Phoenix.Sync.Electric\n      end\n    end\n\nAs for the Plug integration, include the configuration at runtime\nwithin the `Application.start/2` callback.\n\n    # in application.ex\n    def start(_type, _args) do\n      children = [\n        # ...\n        {MyAppWeb.Endpoint, phoenix_sync: Phoenix.Sync.plug_opts()}\n      ]\n\n      Supervisor.start_link(children, strategy: :one_for_one, name: MyApp.Supervisor)\n    end","ref":"Phoenix.Sync.Electric.html#module-phoenix-integration"},{"type":"module","title":"Phoenix.Sync.LiveView","doc":"Swap out `Phoenix.LiveView.stream/3` for `Phoenix.Sync.LiveView.sync_stream/4` to\nautomatically keep a LiveView up-to-date with the state of your Postgres database:\n\n```elixir\ndefmodule MyWeb.MyLive do\n  use Phoenix.LiveView\n  import Phoenix.Sync.LiveView\n\n  def mount(_params, _session, socket) do\n    {:ok, sync_stream(socket, :todos, Todos.Todo)}\n  end\n\n  def handle_info({:sync, event}, socket) do\n    {:noreply, sync_stream_update(socket, event)}\n  end\nend\n```","ref":"Phoenix.Sync.LiveView.html"},{"type":"opaque","title":"Phoenix.Sync.LiveView.component_event/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:component_event/0"},{"type":"type","title":"Phoenix.Sync.LiveView.event/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:event/0"},{"type":"opaque","title":"Phoenix.Sync.LiveView.replication_event/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:replication_event/0"},{"type":"opaque","title":"Phoenix.Sync.LiveView.root_event/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:root_event/0"},{"type":"type","title":"Phoenix.Sync.LiveView.state_event/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:state_event/0"},{"type":"type","title":"Phoenix.Sync.LiveView.stream_option/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:stream_option/0"},{"type":"type","title":"Phoenix.Sync.LiveView.stream_options/0","doc":"","ref":"Phoenix.Sync.LiveView.html#t:stream_options/0"},{"type":"function","title":"Phoenix.Sync.LiveView.sync_stream/4","doc":"Maintains a LiveView stream from the given Ecto query.\n\n- `name` The name to use for the LiveView stream.\n- `query` An [`Ecto`](`Ecto`) query that represents the data to stream from the database.\n\nFor example:\n\n    def mount(_params, _session, socket) do\n      socket =\n        Phoenix.Sync.LiveView.sync_stream(\n          socket,\n          :admins,\n          from(u in Users, where: u.admin == true)\n        )\n      {:ok, socket}\n    end\n\nThis will subscribe to the configured Electric server and keep the list of\n`:admins` in sync with the database via a `Phoenix.LiveView` stream.\n\nUpdates will be delivered to the view via messages to the LiveView process.\n\nTo handle these you need to add a `handle_info/2` implementation that receives these:\n\n    def handle_info({:sync, event}, socket) do\n      {:noreply, Phoenix.Sync.LiveView.sync_stream_update(socket, event)}\n    end\n\nSee the docs for `Phoenix.LiveView.stream/4` for details on using LiveView streams.","ref":"Phoenix.Sync.LiveView.html#sync_stream/4"},{"type":"function","title":"Lifecycle Events - Phoenix.Sync.LiveView.sync_stream/4","doc":"Most `{:sync, event}` messages are opaque and should be passed directly\nto the `sync_stream_update/3` function, but there are two events that are\noutside Electric's replication protocol and designed to be useful in the\nLiveView component.\n\n- `{:sync, {stream_name, :loaded}}` - sent when the Electric event stream has passed\n  from initial state to update mode.\n\n  This event is useful to show the stream component after\n  the initial sync. Because of the streaming nature of Electric Shapes, the\n  intitial sync can cause flickering as items are added, removed and updated.\n\n  E.g.:\n\n      # in the LiveView component\n      def handle_info({:sync, {_name, :live}}, socket) do\n        {:noreply, assign(socket, :show_stream, true)}\n      end\n\n      # in the template\n       \n         \n          <%= item.value %>\n         \n       \n\n- `{:sync, {stream_name, :live}}` - sent when the Electric stream is in\n  `live` mode, that is the initial state has loaded and the client is\n  up-to-date with the database and is long-polling for new events from the\n  Electric server.\n\nIf your app doesn't need this extra information, then you can ignore them and\njust have a catch-all callback:\n\n    def handle_info({:sync, event}, socket) do\n      {:noreply, Phoenix.Sync.LiveView.sync_stream_update(socket, event)}\n    end\n\n`Phoenix.Sync.LiveView.sync_stream_update` will just ignore the\nlifecycle events.","ref":"Phoenix.Sync.LiveView.html#sync_stream/4-lifecycle-events"},{"type":"function","title":"Sub-components - Phoenix.Sync.LiveView.sync_stream/4","doc":"If you register your Electric stream in a sub-component you will still\nreceive Electric messages in the LiveView's root/parent process.\n\n`Phoenix.Sync` handles this for you by encapsulating component messages\nso it can correctly forward on the event to the component.\n\nSo in the parent `LiveView` process you handle the `:sync` messages as\nabove:\n\n    defmodule MyLiveView do\n      use Phoenix.LiveView\n\n      def render(assigns) do\n        ~H\"\"\"\n         \n          <.live_component id=\"my_component\" module={MyComponent} />\n         \n        \"\"\"\n      end\n\n      # We setup the Electric sync_stream in the component but update messages will\n      # be sent to the parent process.\n      def handle_info({:sync, event}, socket) do\n        {:noreply, Phoenix.Sync.LiveView.sync_stream_update(socket, event)}\n      end\n    end\n\nIn the component you must handle these events in the\n`c:Phoenix.LiveComponent.update/2` callback:\n\n    defmodule MyComponent do\n      use Phoenix.LiveComponent\n\n      def render(assigns) do\n        ~H\"\"\"\n         \n           \n            <%= user.name %>\n           \n         \n        \"\"\"\n      end\n\n      # Equivalent to the `handle_info({:sync, {stream_name, :live}}, socket)` callback\n      # in the parent LiveView.\n      def update(%{sync: {_stream_name, :live}}, socket) do\n        {:ok, socket}\n      end\n\n      # Equivalent to the `handle_info({:sync, event}, socket)` callback\n      # in the parent LiveView.\n      def update(%{sync: event}, socket) do\n        {:ok, Phoenix.Sync.LiveView.sync_stream_update(socket, event)}\n      end\n\n      def update(assigns, socket) do\n        {:ok, Phoenix.Sync.LiveView.sync_stream(socket, :users, User)}\n      end\n    end","ref":"Phoenix.Sync.LiveView.html#sync_stream/4-sub-components"},{"type":"function","title":"Phoenix.Sync.LiveView.sync_stream_update/3","doc":"Handle Electric events within a LiveView.\n\n    def handle_info({:sync, event}, socket) do\n      {:noreply, Phoenix.Sync.LiveView.sync_stream_update(socket, event, at: 0)}\n    end\n\nThe `opts` are passed to the `Phoenix.LiveView.stream_insert/4` call.","ref":"Phoenix.Sync.LiveView.html#sync_stream_update/3"},{"type":"module","title":"Phoenix.Sync.Router","doc":"Provides router macros to simplify the exposing of Electric shape streams\nwithin your Phoenix or Plug application.","ref":"Phoenix.Sync.Router.html"},{"type":"module","title":"Phoenix Integration - Phoenix.Sync.Router","doc":"When using within a Phoenix application, you should just import the macros\ndefined here in your `Phoenix.Router` module:\n\n    defmodule MyAppWeb.Router do\n      use Phoenix.Router\n\n      import Elixir.Phoenix.Sync.Router\n\n      scope \"/shapes\" do\n        sync \"/all-todos\", MyApp.Todos.Todo\n\n        sync \"/pending-todos\", MyApp.Todos.Todo,\n          where: \"completed = false\"\n      end\n    end","ref":"Phoenix.Sync.Router.html#module-phoenix-integration"},{"type":"module","title":"Plug Integration - Phoenix.Sync.Router","doc":"Within your `Plug.Router` module, `use Elixir.Phoenix.Sync.Router` and then \nadd your `sync` routes:\n\n    defmodule MyApp.Plug.Router do\n      use Plug.Router, copy_opts_to_assign: :options\n      use Elixir.Phoenix.Sync.Router\n\n      plug :match\n      plug :dispatch\n\n      sync \"/shapes/all-todos\", MyApp.Todos.Todo\n\n      sync \"/shapes/pending-todos\", MyApp.Todos.Todo,\n        where: \"completed = false\"\n    end\n\nYou **must** use the `copy_opts_to_assign` option in `Plug.Router` in order\nfor the `sync` macro to get the configuration defined in your\n`application.ex` [`start/2`](`c:Application.start/2`) callback.","ref":"Phoenix.Sync.Router.html#module-plug-integration"},{"type":"macro","title":"Phoenix.Sync.Router.sync/2","doc":"Defines a synchronization route for streaming Electric shapes.\n\nThe shape can be defined in several ways:","ref":"Phoenix.Sync.Router.html#sync/2"},{"type":"macro","title":"Using Ecto Schemas - Phoenix.Sync.Router.sync/2","doc":"Defines a synchronization route for streaming Electric shapes using an Ecto schema.\n\n    sync \"/all-todos\", MyApp.Todo\n\nNote: Only Ecto schema modules are supported as direct arguments. For Ecto queries,\nuse the `query` option in the third argument or use `Phoenix.Sync.Controller.sync_render/3`.","ref":"Phoenix.Sync.Router.html#sync/2-using-ecto-schemas"},{"type":"macro","title":"Using Ecto Schema and `where` clause - Phoenix.Sync.Router.sync/2","doc":"sync \"/incomplete-todos\", MyApp.Todo, where: \"completed = false\"","ref":"Phoenix.Sync.Router.html#sync/2-using-ecto-schema-and-where-clause"},{"type":"macro","title":"Using an explicit `table` - Phoenix.Sync.Router.sync/2","doc":"sync \"/incomplete-todos\", table: \"todos\", where: \"completed = false\"\n\n\nSee [the section on Shape definitions](readme.html#shape-definitions) for\nmore details on keyword-based shapes.","ref":"Phoenix.Sync.Router.html#sync/2-using-an-explicit-table"},{"type":"macro","title":"Phoenix.Sync.Router.sync/3","doc":"Create a synchronization route from an `Ecto.Schema` plus shape options.\n\n    sync \"/my-shape\", MyApp.Todos.Todo,\n      where: \"completed = false\"\n\nSee `sync/2`.","ref":"Phoenix.Sync.Router.html#sync/3"},{"type":"module","title":"Phoenix.Sync.Writer","doc":"Provides [write-path sync](https://electric-sql.com/docs/guides/writes) support for\nPhoenix- or Plug-based apps.\n\nImagine you're building an application on sync. You've used the\n[read-path sync utilities](../../../README.md#read-path-sync) to sync data into the\nfront-end. If the client then changes the data locally, these writes can be batched\nup and sent back to the server.\n\n`Phoenix.Sync.Writer` provides a principled way of ingesting these local writes\nand applying them to Postgres. In a way that works-with and re-uses your existing\nauthorization logic and your existing `Ecto.Schema`s and `Ecto.Changeset` validation\nfunctions.\n\nThis allows you to build instant, offline-capable applications that work with\n[local optimistic state](https://electric-sql.com/docs/guides/writes).","ref":"Phoenix.Sync.Writer.html"},{"type":"module","title":"Usage levels ([low](#module-low-level-usage-diy), [mid](#module-mid-level-usage), [high](#module-high-level-usage)) - Phoenix.Sync.Writer","doc":"You don't need to use `Phoenix.Sync.Writer` to ingest write operations using Phoenix.\nPhoenix already ships with primitives like `Ecto.Multi` and `c:Ecto.Repo.transaction/2`.\nHowever, `Phoenix.Sync.Writer` provides:\n\n- a number of convienience functions that simplify ingesting mutation operations\n- a high-level pipeline that dries up a lot of common boilerplate and allows you to re-use\n  your existing `Plug` and `Ecto.Changeset` logic","ref":"Phoenix.Sync.Writer.html#module-usage-levels-low-module-low-level-usage-diy-mid-module-mid-level-usage-high-module-high-level-usage"},{"type":"module","title":"Low-level usage (DIY) - Phoenix.Sync.Writer","doc":"If you're comfortable parsing, validating and persisting changes yourself then the\nsimplest way to use `Phoenix.Sync.Writer` is to use `txid!/1` within\n`c:Ecto.Repo.transaction/2`:\n\n    {:ok, txid} =\n      MyApp.Repo.transaction(fn ->\n        # ... save your changes to the database ...\n\n        # Return the transaction id.\n        Phoenix.Sync.Writer.txid!(MyApp.Repo)\n      end)\n\nThis returns the database transaction ID that the changes were applied within. This allows\nyou to return it to the client, which can then monitor the read-path sync stream to detect\nwhen the transaction syncs through. At which point the client can discard its local\noptimistic state.\n\nA convienient way of doing this is to parse the request data into a list of\n`Phoenix.Sync.Writer.Operation`s using a `Phoenix.Sync.Writer.Format`.\nYou can then apply the changes yourself by matching on the operation data:\n\n    {:ok, %Transaction{operations: operations}} =\n      Phoenix.Sync.Writer.parse_transaction(\n        my_encoded_txn,\n        format: Phoenix.Sync.Writer.Format.TanstackDB\n      )\n\n    {:ok, txid} =\n      MyApp.Repo.transaction(fn ->\n        Enum.each(txn.operations, fn\n          %{operation: :insert, relation: [_, \"todos\"], change: change} ->\n            # insert a Todo\n          %{operation: :update, relation: [_, \"todos\"], data: data, change: change} ->\n            # update a Todo\n          %{operation: :delete, relation: [_, \"todos\"], data: data} ->\n            # for example, if you don't want to allow deletes...\n            raise \"invalid delete\"\n        end)\n\n        Phoenix.Sync.Writer.txid!(MyApp.Repo)\n      end, timeout: 60_000)","ref":"Phoenix.Sync.Writer.html#module-low-level-usage-diy"},{"type":"module","title":"Mid-level usage - Phoenix.Sync.Writer","doc":"The pattern above is wrapped-up into the more convienient `transact/4` function.\nThis abstracts the parsing and txid details whilst still allowing you to handle\nand apply mutation operations yourself:\n\n    {:ok, txid} =\n      Phoenix.Sync.Writer.transact(\n        my_encoded_txn,\n        MyApp.Repo,\n        fn\n          %{operation: :insert, relation: [_, \"todos\"], change: change} ->\n            MyApp.Repo.insert(...)\n          %{operation: :update, relation: [_, \"todos\"], data: data, change: change} ->\n            MyApp.Repo.update(Ecto.Changeset.cast(...))\n          %{operation: :delete, relation: [_, \"todos\"], data: data} ->\n            # we don't allow deletes...\n            {:error, \"invalid delete\"}\n        end,\n        format: Phoenix.Sync.Writer.Format.TanstackDB,\n        timeout: 60_000\n      )\n\nHowever, with larger applications, this flexibility can become tiresome as you end up\nrepeating boilerplate and defining your own pipeline to authorize, validate and apply\nchanges with the right error handling and return values.","ref":"Phoenix.Sync.Writer.html#module-mid-level-usage"},{"type":"module","title":"High-level usage - Phoenix.Sync.Writer","doc":"To avoid this, `Phoenix.Sync.Writer` provides a higer level pipeline that dries up\nthe boilerplate, whilst still allowing flexibility and extensibility. You create an\ningest pipeline by instantiating a `Phoenix.Sync.Writer` instance and piping into\n`allow/3` and `apply/4` calls:\n\n    {:ok, txid, _changes} =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todo)\n      |> Phoenix.Sync.Writer.allow(MyApp.OtherSchema)\n      |> Phoenix.Sync.Writer.apply(transaction, Repo, format: MyApp.MutationFormat)\n\nOr, instead of `apply/4` you can use seperate calls to `ingest/3` and then `transaction/2`.\nThis allows you to ingest multiple formats, for example:\n\n    {:ok, txid} =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todo)\n      |> Phoenix.Sync.Writer.ingest(changes, format: MyApp.MutationFormat)\n      |> Phoenix.Sync.Writer.ingest(other_changes, parser: &MyApp.MutationFormat.parse_other/1)\n      |> Phoenix.Sync.Writer.ingest(more_changes, parser: {MyApp.MutationFormat, :parse_more, []})\n      |> Phoenix.Sync.Writer.transaction(MyApp.Repo)\n\nAnd at any point you can drop down / eject out to the underlying `Ecto.Multi` using\n`to_multi/1` or `to_multi/3`:\n\n    multi =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todo)\n      |> Phoenix.Sync.Writer.to_multi(changes, format: MyApp.MutationFormat)\n\n    # ... do anything you like with the multi ...\n\n    {:ok, changes} = Repo.transaction(multi)\n    {:ok, txid} = Phoenix.Sync.Writer.txid(changes)","ref":"Phoenix.Sync.Writer.html#module-high-level-usage"},{"type":"module","title":"Controller example - Phoenix.Sync.Writer","doc":"For example, take a project management app that's using\n[@TanStack/db](https://github.com/TanStack/db) to batch up local\noptimistic writes and POST them to the `Phoenix.Controller` below:\n\n    defmodule MutationController do\n      use Phoenix.Controller, formats: [:json]\n\n      alias Phoenix.Sync.Writer\n      alias Phoenix.Sync.Writer.Format\n\n      def mutate(conn, %{\"transaction\" => transaction} = _params) do\n        user_id = conn.assigns.user_id\n\n        {:ok, txid, _changes} =\n          Phoenix.Sync.Writer.new()\n          |> Phoenix.Sync.Writer.allow(\n            Projects.Project,\n            check: reject_invalid_params/2,\n            load: &Projects.load_for_user(&1, user_id),\n            validate: &Projects.Project.changeset/2\n          )\n          |> Phoenix.Sync.Writer.allow(\n            Projects.Issue,\n            # Use the sensible defaults:\n            # validate: Projects.Issue.changeset/2\n            # etc.\n          )\n          |> Phoenix.Sync.Writer.apply(\n            transaction,\n            Repo,\n            format: Format.TanstackDB\n          )\n\n        render(conn, :mutations, txid: txid)\n      end\n    end\n\nThe controller constructs a `Phoenix.Sync.Writer` instance and pipes\nit through a series of `allow/3` calls, registering functions against\n`Ecto.Schema`s (in this case `Projects.Project` and `Projects.Issue`) to\nvalidate and authorize each of these mutation operations before applying\nthem as a single transaction.\n\nThis controller can become a single, unified entry point for ingesting writes\ninto your application. You can extend the pipeline with `allow/3` calls for\nevery schema that you'd like to be able to ingest changes to.\n\nThe [`check`, `load`, `validate`, etc. callbacks](#callbacks) to the allow\nfunction are designed to allow you to re-use your authorization and validation\nlogic from your existing `Plug` middleware and `Ecto.Changeset` functions.\n\n> #### Warning {: .warning}\n>\n> The mutation operations received from clients MUST be considered as **untrusted**.\n>\n> Though the HTTP operation that uploaded them will have been authenticated and\n> authorized by your existing Plug middleware as usual, the actual content of the\n> request that is turned into writes against your database needs to be validated\n> very carefully against the privileges of the current user.\n>\n> That's what `Phoenix.Sync.Writer` is for: specifying which resources can be\n> updated and registering functions to authorize and validate the mutation payload.","ref":"Phoenix.Sync.Writer.html#module-controller-example"},{"type":"module","title":"Transactions - Phoenix.Sync.Writer","doc":"The `txid` in the return value from `apply/4` and `txid/1` / `txid!/1` allows the\nPostgres transaction ID to be returned to the client in the response data.\n\nThis allows clients to monitor the read-path sync stream and match on the\narrival of the same transaction id. When the client receives this transaction id\nback through its sync stream, it knows that it can discard the local optimistic\nstate for that transaction. (This is a more robust way of managing optimistic state\nthan just matching on instance IDs, as it allows for local changes to be rebased\non concurrent changes to the same date from other users).\n\n`Phoenix.Sync.Writer` uses `Ecto.Multi`'s transaction update mechanism\nunder the hood, which means that either all the operations in a client\ntransaction are accepted or none are. See `to_multi/1` for how you can hook\ninto the `Ecto.Multi` after applying your change data.\n\n> #### Compatibility {: .info}\n>\n> `Phoenix.Sync.Writer` can only return transaction ids when connecting to\n> a Postgres database (a repo with `adapter: Ecto.Adapters.Postgres`). You can\n> use this module for other databases, but the returned txid will be `nil`.","ref":"Phoenix.Sync.Writer.html#module-transactions"},{"type":"module","title":"Client Libraries - Phoenix.Sync.Writer","doc":"`Phoenix.Sync.Writer` is not coupled to any particular client-side implementation.\nSee Electric's [write pattern guides and example code](https://electric-sql.com/docs/guides/writes)\nfor implementation strategies and examples.\n\nInstead, `Phoenix.Sync.Writer` provides an adapter pattern where you can register\na `format` adapter or `parser` function to parse the expected payload format from a client side library\ninto the struct that `Phoenix.Sync.Writer` expects.\n\nThe currently supported format adapters are:\n\n- [TanStack/db](https://github.com/TanStack/db) \"A reactive client store for\n  building super fast apps on sync\"\n\n  Integration:\n\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.ingest(mutation_data, format: Phoenix.Sync.Writer.Format.TanstackDB)\n      |> Phoenix.Sync.Writer.transaction(Repo)","ref":"Phoenix.Sync.Writer.html#module-client-libraries"},{"type":"module","title":"Usage - Phoenix.Sync.Writer","doc":"Much as every controller action must be authenticated, authorized and validated\nto prevent users writing invalid data or data that they do not have permission\nto modify, mutations **MUST** be validated for both correctness (are the given\nvalues valid?) and permissions (is the current user allowed to apply the given\nmutation?).\n\nThis dual verification -- of data and permissions -- is performed by a pipeline\nof application-defined callbacks for every model that you allow writes to:\n\n- [`check`](#module-check) - a function that performs a \"pre-flight\"\n  sanity check of the user-provided data in the mutation; this should just\n  validate the data and not usually hit the database; checks are performed on\n  all operations in a transaction before proceeding to the next steps in the\n  pipeline; this allows for fast rejection of invalid data before performing\n  more expensive operations\n- [`load`](#module-load) - a function that takes the original data\n  and returns the existing model from the database, if it exists, for an update\n  or delete operation\n- [`validate`](#module-validate) - create and validate an `Ecto.Changeset`\n  from the source data and mutation changes; this is intended to be compatible with\n  using existing schema changeset functions; note that, as per any changeset function,\n  the validate function can perform both authorization and validation\n- [`pre_apply` and `post_apply`](#module-pre_apply-and-post_apply) - add\n  arbitrary `Ecto.Multi` operations to the transaction based on the current operation\n\nSee `apply/4` and the [Callbacks](#module-callbacks) for how the transaction is\nprocessed internally and how best to use these callback functions to express your\napp's authorization and validation requirements.\n\nCalling `new/0` creates an empty writer configuration with the given mutation\nparser. But this alone does not permit any mutations. In order to allow writes\nfrom clients you must call `allow/3` with a schema module and some callback functions.\n\n    # create an empty writer configuration\n    writer = Phoenix.Sync.Writer.new()\n\n    # allow writes to the `Todos.Todo` table\n    # using `Todos.check_mutation/1` to validate mutation data before\n    # touching the database\n    writer = Phoenix.Sync.Writer.allow(writer, Todos.Todo, check: &Todos.check_mutation/1)\n\nIf the table name on the client differs from the Postgres table, then you can\nadd a `table` option that specifies the client table name that this `allow/3`\ncall applies to:\n\n    # `client_todos` is the name of the `todos` table on the clients\n    writer =\n      Phoenix.Sync.Writer.allow(\n        writer,\n        Todos.Todo,\n        validate: &Todos.validate_mutation/2,\n        table: \"client_todos\"\n      )","ref":"Phoenix.Sync.Writer.html#module-usage"},{"type":"module","title":"Callbacks - Phoenix.Sync.Writer","doc":"","ref":"Phoenix.Sync.Writer.html#module-callbacks"},{"type":"module","title":"Check - Phoenix.Sync.Writer","doc":"The `check` option should be a 1-arity function whose purpose is to test\nthe mutation data against the authorization rules for the application and\nmodel before attempting any database access.\n\nIf the changes are valid then it should return `:ok` or `{:error, reason}` if\nthey're invalid.\n\nIf any of the changes fail the auth test, then the entire transaction will be\nrejected.\n\nThis is the first line of defence against malicious writes as it provides a\nquick check of the data from the clients before any reads or writes to the\ndatabase.\n\nNote that the writer pipeline checks all the operations before proceeding to\nload, validate and apply each operation in turn.\n\n    def check(%Phoenix.Sync.Writer.Operation{} = operation) do\n      # :ok or {:error, \"...\"}\n    end","ref":"Phoenix.Sync.Writer.html#module-check"},{"type":"module","title":"Load - Phoenix.Sync.Writer","doc":"The `load` callback takes the `data` in `update` or `delete` mutations (i.e.:\nthe original data before changes), and uses it to retrieve the original\n`Ecto.Struct` model from the database.\n\nIt can be a 1- or 2-arity function. The 1-arity version receives just the\n`data` parameters. The 2-arity version receives the `Ecto.Repo` that the\ntransaction is being applied to and the `data` parameters.\n\n    # 1-arity version\n    def load(%{\"column\" => \"value\"} = data) do\n      # Repo.get(...)\n    end\n\n    # 2-arity version\n    def load(repo, %{\"column\" => \"value\"} = data) do\n      # repo.get(...)\n    end\n\nIf not provided defaults to using `c:Ecto.Repo.get_by/3` using the primary\nkey(s) defined on the model.\n\nFor `insert` operations this load function is not used. Instead, the original\nstruct is created by calling the `__struct__/0` function on the `Ecto.Schema`\nmodule.","ref":"Phoenix.Sync.Writer.html#module-load"},{"type":"module","title":"Validate - Phoenix.Sync.Writer","doc":"The `validate` callback performs the usual role of a changeset function: to\nvalidate the changes against the model's data constraints using the functions\nin `Ecto.Changeset`.\n\nIt should return an `Ecto.Changeset` instance (or possibly the original\nschema struct in the case of `delete`s). If any of the transaction's\nchangeset's are marked as invalid, then the entire transaction is aborted.\n\nIf not specified, the `validate` function is defaulted to the schema model's\nstandard `changeset/2` function if available.\n\nThe callback can be either a 2- or 3-arity function.\n\nThe 2-arity version will receive the `Ecto.Schema` struct returned from the\n`load` function and the mutation changes. The 3-arity version will receive\nthe `load`ed struct, the changes and the operation.\n\n    # 2-arity version\n    def changeset(%Todo{} = data, %{} = changes) do\n      data\n      |> Ecto.Changeset.cast(changes, [:title, :completed])\n      |> Ecto.Changeset.validate_required(changes, [:title])\n    end\n\n    # 3-arity version\n    def changeset(%Todo{} = data, %{} = changes, operation)\n        when operation in [:insert, :update, :delete] do\n      # ...\n    end\n\n#### Primary keys\n\nWhether the params for insert operations contains a value for the new primary\nkey is application specific. It's certainly not required if you have declared your\n`Ecto.Schema` model with a primary key set to `autogenerate: true`.\n\nIt's worth noting that if you are accepting primary key values as part of\nyour `insert` changes, then you should use UUID primary keys for your models\nto prevent conflicts.","ref":"Phoenix.Sync.Writer.html#module-validate"},{"type":"module","title":"pre_apply and post_apply - Phoenix.Sync.Writer","doc":"These callbacks, run before or after the actual `insert`, `update` or\n`delete` operation allow for the addition of side effects to the transaction.\n\nThey are passed an empty `Ecto.Multi` struct and which is then\n[merged](`Ecto.Multi.merge/2`) into the writer's transaction.\n\nThey also allow for more validation/authorization steps as any operation\nwithin the callback that returns an \"invalid\" operation will abort the entire\ntransaction.\n\n    def pre_or_post_apply(%Ecto.Multi{} = multi, %Ecto.Changeset{} = change, %Phoenix.Sync.Writer.Context{} = context) do\n      multi\n      # add some side-effects\n      # |> Ecto.Multi.run(Phoenix.Sync.Writer.operation_name(context, :image), fn _changes ->\n      #   with :ok <- File.write(image.name, image.contents) do\n      #    {:ok, nil}\n      #   end\n      # end)\n      #\n      # validate the current transaction and abort using an {:error, value} tuple\n      # |> Ecto.Multi.run(Phoenix.Sync.Writer.operation_name(context, :my_validation), fn _changes ->\n      #   {:error, \"reject entire transaction\"}\n      # end)\n    end\n\nNote the use of `operation_name/2` when adding operations. Every name in the\nfinal `Ecto.Multi` struct must be unique, `operation_name/2` generates names\nthat are guaranteed to be unique to the current operation and callback.","ref":"Phoenix.Sync.Writer.html#module-pre_apply-and-post_apply"},{"type":"module","title":"Per-operation callbacks - Phoenix.Sync.Writer","doc":"If you want to re-use an existing function on a per-operation basis, then in\nyour write configuration you can define both top-level and per operation\ncallbacks:\n\n    Phoenix.Sync.Writer.allow(\n      Todos.Todo,\n      load: &Todos.fetch_for_user(&1, user_id),\n      check: &Todos.check_mutation(&1, &2, user_id),\n      validate: &Todos.Todo.changeset/2,\n      update: [\n        # for inserts and deletes, use &Todos.Todo.changeset/2 but for updates\n        # use this function\n        validate: &Todos.Todo.update_changeset/2,\n        pre_apply: &Todos.pre_apply_update_todo/3\n      ],\n      insert: [\n        # optional validate, pre_apply and post_apply\n        # overrides for insert operations\n      ],\n      delete: [\n        # optional validate, pre_apply and post_apply\n        # overrides for delete operations\n      ],\n    )","ref":"Phoenix.Sync.Writer.html#module-per-operation-callbacks"},{"type":"module","title":"End-to-end usage - Phoenix.Sync.Writer","doc":"The combination of the `check`, `load`, `validate`, `pre_apply` and\n`post_apply` functions can be composed to provide strong guarantees of\nvalidity.\n\nThe aim is to provide an interface as similar to that used in controller\nfunctions as possible.\n\nHere we show an example controller module that allows updating of `Todo`s via\na standard `HTTP PUT` update handler and also via `HTTP POST`s to the\n`mutation` handler which applies optimistic writes via this module.\n\nWe use the `load` function to validate the ownership of the original `Todo`\nby looking up the data using both the `id` and the `user_id`. This makes it\nimpossible for user `a` to update todos belonging to user `b`.\n\n    defmodule MyController do\n      use Phoenix.Controller, formats: [:html, :json]\n\n      alias Phoenix.Sync.Writer\n\n      # The standard HTTP PUT update handler\n      def update(conn, %{\"todo\" => todo_params}) do\n        user_id = conn.assigns.user_id\n\n        with {:ok, todo} <- fetch_for_user(params, user_id),\n             {:ok, params} <- validate_params(todo, todo_params, user_id),\n             {:ok, updated_todo} <- Todos.update(todo, params) do\n          redirect(conn, to: ~p\"/todos/#{updated_todo.id}\")\n        end\n      end\n\n      # The HTTP POST mutations handler which receives JSON data\n      def mutations(conn, %{\"transaction\" => transaction} = _params) do\n        user_id = conn.assigns.user_id\n\n        {:ok, txid, _changes} =\n          Writer.new()\n          |> Writer.allow(\n            Todos.Todo,\n            check: &validate_mutation(&1, user_id),\n            load: &fetch_for_user(&1, user_id),\n          )\n          |> Writer.apply(transaction, Repo, format: Writer.Format.TanstackDB)\n\n        render(conn, :mutations, txid: txid)\n      end\n\n      # Included here for completeness but in a real app would be a\n      # public function in the Todos context.\n      # Because we're validating the ownership of the Todo here we add an\n      # extra layer of auth checks, preventing one user from modifying\n      # the Todos of another.\n      defp fetch_for_user(%{\"id\" => id}, user_id) do\n        from(t in Todos.Todo, where: t.id == ^id  and t.user_id == ^user_id)\n        |> Repo.one()\n      end\n\n      defp validate_mutation(%Writer.Operation{} = op, user_id) do\n        with :ok <- validate_params(op.data, user_id) do\n          validate_params(op.changes, user_id)\n        end\n      end\n\n      defp validate_params(%{\"user_id\" => user_id}, user_id), do: :ok\n      defp validate_params(%{} = _params, _user_id), do: {:error, \"invalid user_id\"}\n    end\n\nBecause `Phoenix.Sync.Write` leverages `Ecto.Multi` to do the work of\napplying changes and managing errors, you're also free to extend the actions\nthat are performed with every transaction using `pre_apply` and `post_apply`\ncallbacks configured per-table or per-table per-action (insert, update,\ndelete). See `allow/3` for more information on the configuration options\nfor each table.\n\nThe result of `to_multi/1` or `to_multi/3` is an `Ecto.Multi` instance so you can also just\nappend operations using the normal `Ecto.Multi` functions:\n\n    {:ok, txid, _changes} =\n      Writer.new()\n      |> Writer.allow(Todo, ...)\n      |> Writer.to_multi(transaction, parser: &my_transaction_parser/1)\n      |> Ecto.Multi.insert(:my_action, %Event{})\n      |> Writer.transaction(Repo)","ref":"Phoenix.Sync.Writer.html#module-end-to-end-usage"},{"type":"function","title":"Phoenix.Sync.Writer.allow/3","doc":"Allow writes to the given `Ecto.Schema`.\n\nOnly tables specified in calls to `allow/3` will be accepted by later calls\nto `transaction/3`. Any changes to tables not explicitly defined by `allow/3` calls\nwill be rejected and cause the entire transaction to be rejected.","ref":"Phoenix.Sync.Writer.html#allow/3"},{"type":"function","title":"Examples - Phoenix.Sync.Writer.allow/3","doc":"# allow writes to the Todo table using\n    # `MyApp.Todos.Todo.check_mutation/1` to validate operations\n    Phoenix.Sync.Writer.new()\n    |> Phoenix.Sync.Writer.allow(\n      MyApp.Todos.Todo,\n      check: &MyApp.Todos.check_mutation/1\n    )\n\n    # A more complex configuration adding an `post_apply` callback to inserts\n    # and using a custom query to load the original database value.\n    Phoenix.Sync.Writer.new()\n    |> Phoenix.Sync.Writer.allow(\n      MyApp.Todos..Todo,\n      load: &MyApp.Todos.get_for_mutation/1,\n      check: &MyApp.Todos.check_mutation/1,\n      insert: [\n        post_apply: &MyApp.Todos.post_apply_insert_mutation/3\n      ]\n    )","ref":"Phoenix.Sync.Writer.html#allow/3-examples"},{"type":"function","title":"Supported options - Phoenix.Sync.Writer.allow/3","doc":"* `:table` (`String.t() | [String.t(), ...]`) - Override the table name of the `Ecto.Schema` struct to\n  allow for mapping between table names on the client and within Postgres.\n\n  If you pass just a table name, then any schema prefix in the client tables is ignored, so\n\n      Writer.allow(Todos, table: \"todos\")\n\n  will match client operations for `[\"public\", \"todos\"]` and `[\"application\", \"todos\"]` etc.\n\n  If you provide a 2-element list then the mapping will be exact and only\n  client relations matching the full `[schema, table]` pair will match the\n  given schema.\n\n      Writer.allow(Todos, table: [\"public\", \"todos\"])\n\n  Will match client operations for `[\"public\", \"todos\"]` but\n  **not** `[\"application\", \"todos\"]` etc.\n\n  Defaults to `Model.__schema__(:source)`, or if the Ecto schema\n  module has specified a `namespace` `[Model.__schema__(:prefix),\n  Model.__schema__(:source)]`.\n\n* `:accept` - A list of actions to accept.\n\n  A transaction containing an operation not in the accept list will be rejected.\n\n  Defaults to accepting all operations, `[:insert, :update, :delete]`.\n\n* `:check` (function of arity 1) - A function that validates every %Phoenix.Sync.Writer.Operation{} in the transaction for correctness.\n\n  This is run before any database access is performed and so provides an\n  efficient way to prevent malicious writes without hitting your database.\n\n  Defaults to a function that allows all operations: `fn _ -> :ok end`.\n\n* `:before_all` (function of arity 1) - Run only once (per transaction) after the parsing and `check` callback have\n  completed and before `load` and `validate` functions run.\n\n  Useful for pre-loading data from the database that can be shared across\n  all operation callbacks for all the mutations.\n\n  Arguments:\n\n  - `multi` an `Ecto.Multi` struct\n\n  Return value:\n\n  - `Ecto.Multi` struct with associated data\n\n  Defaults to no callback.\n\n* `:load` - A 1- or 2-arity function that accepts either the mutation\n  operation's data or an `Ecto.Repo` instance and the mutation data and\n  returns the original row from the database.\n\n  Arguments:\n\n  - `repo` the `Ecto.Repo` instance passed to `apply/4` or `transaction/3`\n  - `data` the original operation data\n\n  Valid return values are:\n\n  - `struct()` - an `Ecto.Schema` struct, that must match the\n    module passed to `allow/3`\n  - `{:ok, struct()}` - as above but wrapped in an `:ok` tuple\n  - `nil` - if no row matches the search criteria, or\n  - `{:error, String.t()}` - as `nil` but with a custom error string\n\n  A return value of `nil` or `{:error, reason}` will abort the transaction.\n\n  This function is only used for updates or deletes. For\n  inserts, the `__struct__/0` function defined by `Ecto.Schema` is used to\n  create an empty schema struct.","ref":"Phoenix.Sync.Writer.html#allow/3-supported-options"},{"type":"function","title":"Examples - Phoenix.Sync.Writer.allow/3","doc":"# load from a known Repo\n      load: fn %{\"id\" => id} -> MyApp.Repo.get(Todos.Todo, id)\n\n      # load from the repo passed to `Elixir.Phoenix.Sync.Writer.transaction/2`\n      load: fn repo, %{\"id\" => id} -> repo.get(Todos.Todo, id)\n\n  If not provided defaults to `c:Ecto.Repo.get_by/3` using the\n  table's schema module and its primary keys.\n\n* `:validate` - a 2- or 3-arity function that returns an `Ecto.Changeset` for a given mutation.","ref":"Phoenix.Sync.Writer.html#allow/3-examples"},{"type":"function","title":"Callback params - Phoenix.Sync.Writer.allow/3","doc":"- `data` an Ecto.Schema struct matching the one used when\n    calling `allow/2` returned from the `load` function.\n  - `changes` a map of changes to apply to the `data`.\n  - `operation` (for 3-arity callbacks only) the operation\n    action, one of `:insert`, `:update` or `:delete`\n\n  At absolute minimum, this should call\n  `Ecto.Changeset.cast/3` to validate the proposed data:\n\n      def my_changeset(data, changes, _operation) do\n        Ecto.Changeset.cast(data, changes, @permitted_columns)\n      end\n\n  Defaults to the given model's `changeset/2` function if\n  defined, raises if no changeset function can be found.\n\n* `:pre_apply` (`t:pre_post_func/0`) - an optional callback that allows for the pre-pending of\n  operations to the `Ecto.Multi` representing a mutation transaction.\n\n  If should be a 3-arity function.","ref":"Phoenix.Sync.Writer.html#allow/3-callback-params"},{"type":"function","title":"Arguments - Phoenix.Sync.Writer.allow/3","doc":"- `multi` - an empty `%Ecto.Multi{}` instance that you should apply\n    your actions to\n  - `changeset` - the changeset representing the individual mutation operation\n  - `context` - the current change [context](`Elixir.Phoenix.Sync.Writer.Context`)\n\n  The result should be the `Ecto.Multi` instance which will be\n  [merged](`Ecto.Multi.merge/2`) with the one representing the mutation\n  operation.\n\n  Because every action in an `Ecto.Multi` must have a unique\n  key, we advise using the `operation_name/2` function to generate a unique\n  operation name based on the `context`.\n\n      def pre_apply(multi, changeset, context) do\n        name = Phoenix.Sync.Writer.operation_name(context, :event_insert)\n        Ecto.Multi.insert(multi, name, %Event{todo_id: id})\n      end\n\n  Defaults to no `nil`.\n\n* `:post_apply` (`t:pre_post_func/0`) - an optional callback function that allows for the\n  appending of operations to the `Ecto.Multi` representing a mutation\n  transaction.\n\n  See the docs for `:pre_apply` for the function signature and arguments.\n\n  Defaults to no `nil`.\n\n* `:insert` (`t:operation_opts/0`) - Callbacks for validating and modifying `insert` operations.\n\n  Accepts definitions for the `validate`, `pre_apply` and\n  `post_apply` functions for `insert` operations that will override the\n  top-level equivalents.\n\n  See the documentation for `allow/3`.\n\n  The only difference with these callback functions is that\n  the `action` parameter is redundant and therefore not passed.\n\n  Defaults to `[]`, using the top-level functions for all operations.\n\n  * `:validate` (function of arity 2) - A 2-arity function that returns a changeset for the given mutation data.\n\n    Arguments:\n\n    - `schema` the original `Ecto.Schema` model returned from the `load` function\n    - `changes` a map of changes from the mutation operation\n\n    Return value:\n\n    - an `Ecto.Changeset`\n\n  * `:pre_apply` (`t:pre_post_func/0`) - An optional callback that allows for the pre-pending of operations to the `Ecto.Multi`.\n\n    Arguments and return value  as per the global `pre_apply` callback.\n\n  * `:post_apply` (`t:pre_post_func/0`) - An optional callback that allows for the appending of operations to the `Ecto.Multi`.\n\n    Arguments and return value as per the global `post_apply` callback.\n\n* `:update` (`t:operation_opts/0`) - Callbacks for validating and modifying `update` operations.\n  See the documentation for `insert`.\n\n  * `:validate` (function of arity 2) - A 2-arity function that returns a changeset for the given mutation data.\n\n    Arguments:\n\n    - `schema` the original `Ecto.Schema` model returned from the `load` function\n    - `changes` a map of changes from the mutation operation\n\n    Return value:\n\n    - an `Ecto.Changeset`\n\n  * `:pre_apply` (`t:pre_post_func/0`) - An optional callback that allows for the pre-pending of operations to the `Ecto.Multi`.\n\n    Arguments and return value  as per the global `pre_apply` callback.\n\n  * `:post_apply` (`t:pre_post_func/0`) - An optional callback that allows for the appending of operations to the `Ecto.Multi`.\n\n    Arguments and return value as per the global `post_apply` callback.\n\n* `:delete` (`t:operation_opts/0`) - Callbacks for validating and modifying `delete` operations.\n  See the documentation for `insert`.\n\n  * `:validate`\n\n  * `:pre_apply` (`t:pre_post_func/0`) - An optional callback that allows for the pre-pending of operations to the `Ecto.Multi`.\n\n    Arguments and return value  as per the global `pre_apply` callback.\n\n  * `:post_apply` (`t:pre_post_func/0`) - An optional callback that allows for the appending of operations to the `Ecto.Multi`.\n\n    Arguments and return value as per the global `post_apply` callback.","ref":"Phoenix.Sync.Writer.html#allow/3-arguments"},{"type":"type","title":"Phoenix.Sync.Writer.allow_opts/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:allow_opts/0"},{"type":"function","title":"Phoenix.Sync.Writer.apply/4","doc":"Ingest and write changes to the given repo in a single call.\n\n    Phoenix.Sync.Writer.new()\n    |> Phoenix.Sync.Writer.apply(changes, Repo, parser: &MyFormat.parse/1)\n\nis equivalent to:\n\n    Phoenix.Sync.Writer.new()\n    |> Phoenix.Sync.Writer.ingest(changes, parser: &MyFormat.parse/1)\n    |> Phoenix.Sync.Writer.transaction(Repo)","ref":"Phoenix.Sync.Writer.html#apply/4"},{"type":"type","title":"Phoenix.Sync.Writer.context/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:context/0"},{"type":"type","title":"Phoenix.Sync.Writer.data/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:data/0"},{"type":"function","title":"Phoenix.Sync.Writer.ingest/3","doc":"Add the given changes to the operations that will be applied within a `transaction/3`.\n\nExamples:\n\n    {:ok, txid} =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todo)\n      |> Phoenix.Sync.Writer.ingest(changes, format: MyApp.MutationFormat)\n      |> Phoenix.Sync.Writer.ingest(other_changes, parser: &MyApp.MutationFormat.parse_other/1)\n      |> Phoenix.Sync.Writer.ingest(more_changes, parser: {MyApp.MutationFormat, :parse_more, []})\n      |> Phoenix.Sync.Writer.transaction(MyApp.Repo)\n\nSupported options:\n\n* `:format` ([`Format.t()`](`t:Phoenix.Sync.Writer.Format.t/0`)) - A module implementing the `Phoenix.Sync.Writer.Format`\n  behaviour.\n\n  See `Phoenix.Sync.Writer.Format`.\n\n* `:parser` (`Phoenix.Sync.Writer.Format.parser_fun() | mfa()`) - A function that parses some input data and returns a\n  [`%Transaction{}`](`Phoenix.Sync.Writer.Transaction`) struct or an error.\n  See `c:Phoenix.Sync.Writer.Format.parse_transaction/1`.","ref":"Phoenix.Sync.Writer.html#ingest/3"},{"type":"type","title":"Phoenix.Sync.Writer.ingest_change/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:ingest_change/0"},{"type":"type","title":"Phoenix.Sync.Writer.mutation/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:mutation/0"},{"type":"function","title":"Phoenix.Sync.Writer.new/0","doc":"Create a new empty writer.\n\nEmpty writers will reject writes to any tables. You should configure writes\nto the permitted tables by calling `allow/3`.","ref":"Phoenix.Sync.Writer.html#new/0"},{"type":"type","title":"Phoenix.Sync.Writer.operation/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:operation/0"},{"type":"function","title":"Phoenix.Sync.Writer.operation_name/1","doc":"Return a unique operation name for use in `pre_apply` or `post_apply` callbacks.\n\n`Ecto.Multi` requires that all operation names be unique within a\ntransaction. This function gives you a simple way to generate a name for your\nown operations that is guarateed not to conflict with any other.\n\nExample:\n\n    Phoenix.Sync.Writer.new()\n    |> Phoenix.Sync.Writer.allow(\n      MyModel,\n      pre_apply: fn multi, changeset, context ->\n        name = Phoenix.Sync.Writer.operation_name(context)\n        Ecto.Multi.insert(multi, name, AuditEvent.for_changeset(changeset))\n      end\n    )","ref":"Phoenix.Sync.Writer.html#operation_name/1"},{"type":"function","title":"Phoenix.Sync.Writer.operation_name/2","doc":"Like `operation_name/1` but allows for a custom label.","ref":"Phoenix.Sync.Writer.html#operation_name/2"},{"type":"type","title":"Phoenix.Sync.Writer.operation_opts/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:operation_opts/0"},{"type":"type","title":"Phoenix.Sync.Writer.parse_opts/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:parse_opts/0"},{"type":"function","title":"Phoenix.Sync.Writer.parse_transaction/2","doc":"Use the parser configured in the given [`Writer`](`Phoenix.Sync.Writer`)\ninstance to decode the given transaction data.\n\nThis can be used to handle mutation operations explicitly:\n\n    {:ok, txn} = Phoenix.Sync.Writer.parse_transaction(my_json_tx_data, format: Phoenix.Sync.Writer.Format.TanstackDB)\n\n    {:ok, txid} =\n      Repo.transaction(fn ->\n        Enum.each(txn.operations, fn operation ->\n          # do something wih the given operation\n          # raise if something is wrong...\n        end)\n        # return the transaction id\n        Phoenix.Sync.Writer.txid!(Repo)\n      end)","ref":"Phoenix.Sync.Writer.html#parse_transaction/2"},{"type":"type","title":"Phoenix.Sync.Writer.pre_post_func/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:pre_post_func/0"},{"type":"type","title":"Phoenix.Sync.Writer.repo_transaction_opts/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:repo_transaction_opts/0"},{"type":"type","title":"Phoenix.Sync.Writer.schema_config/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:schema_config/0"},{"type":"type","title":"Phoenix.Sync.Writer.t/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:t/0"},{"type":"function","title":"Phoenix.Sync.Writer.to_multi/1","doc":"Given a writer configuration created using `allow/3` translate the list of\nmutations into an `Ecto.Multi` operation.\n\nExample:\n\n    %Ecto.Multi{} = multi =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todos.Todo, check: &my_check_function/1)\n      |> Phoenix.Sync.Writer.allow(MyApp.Options.Option, check: &my_check_function/1)\n      |> Phoenix.Sync.Writer.ingest(changes, format: Phoenix.Sync.Writer.Format.TanstackDB)\n      |> Phoenix.Sync.Writer.to_multi()\n\nIf you want to add extra operations to the mutation transaction, beyond those\napplied by any `pre_apply` or `post_apply` callbacks in your mutation config then use\nthe functions in `Ecto.Multi` to do those as normal.\n\nUse `transaction/3` to apply the changes to the database and return the\ntransaction id.\n\n`to_multi/1` builds an `Ecto.Multi` struct containing the operations required to\nwrite the mutation operations to the database.\n\nThe order of operation is:","ref":"Phoenix.Sync.Writer.html#to_multi/1"},{"type":"function","title":"1. Parse - Phoenix.Sync.Writer.to_multi/1","doc":"The transaction data is parsed, using either the `format` or the `parser` function\nsupplied in `ingest/3`.","ref":"Phoenix.Sync.Writer.html#to_multi/1-1-parse"},{"type":"function","title":"2. Check - Phoenix.Sync.Writer.to_multi/1","doc":"The user input data in each operation in the transaction is tested for validity\nvia the `check` function.\n\nAt this point no database operations have taken place. Errors at the parse or\n`check` stage result in an early exit. The purpose of the `check` callback is\nsanity check the incoming mutation data against basic sanitization rules, much\nas you would do with `Plug` middleware and controller params pattern matching.\n\nNow that we have a list of validated mutation operations, the next step is:","ref":"Phoenix.Sync.Writer.html#to_multi/1-2-check"},{"type":"function","title":"3. Before-all - Phoenix.Sync.Writer.to_multi/1","doc":"Perform any actions defined in the `before_all` callback.\n\nThis only happens once per transaction, the first time the model owning the\ncallback is included in the operation list.\n\nThe following actions happen once per operation in the transaction:","ref":"Phoenix.Sync.Writer.html#to_multi/1-3-before-all"},{"type":"function","title":"4. Load - Phoenix.Sync.Writer.to_multi/1","doc":"The `load` function is called to retrieve the source row from the database\n(for `update` and `delete` operations), or the schema's `__struct__/0`\nfunction is called to instantiate an empty struct (`insert`).","ref":"Phoenix.Sync.Writer.html#to_multi/1-4-load"},{"type":"function","title":"5. Validate - Phoenix.Sync.Writer.to_multi/1","doc":"The `validate` function is called with the result of the `load` function\nand the operation's changes.","ref":"Phoenix.Sync.Writer.html#to_multi/1-5-validate"},{"type":"function","title":"6. Pre-apply - Phoenix.Sync.Writer.to_multi/1","doc":"The `pre_apply` callback is called with a `multi` instance, the result of the\n`validate` function and the current `Context`. The result is\n[merged](`Ecto.Multi.merge/2`) into the transaction's ongoing `Ecto.Multi`.","ref":"Phoenix.Sync.Writer.html#to_multi/1-6-pre-apply"},{"type":"function","title":"7. Apply - Phoenix.Sync.Writer.to_multi/1","doc":"The actual operation is applied to the database using one of\n`Ecto.Multi.insert/4`, `Ecto.Multi.update/4` or `Ecto.Multi.delete/4`, and","ref":"Phoenix.Sync.Writer.html#to_multi/1-7-apply"},{"type":"function","title":"8. Post-apply - Phoenix.Sync.Writer.to_multi/1","doc":"Finally the `post_apply` callback is called.\n\nAny error in any of these stages will abort the entire transaction and leave\nyour database untouched.","ref":"Phoenix.Sync.Writer.html#to_multi/1-8-post-apply"},{"type":"function","title":"Phoenix.Sync.Writer.to_multi/3","doc":"Ingest changes and map them into an `Ecto.Multi` instance ready to apply\nusing `Phoenix.Sync.Writer.transaction/3` or `c:Ecto.Repo.transaction/2`.\n\nThis is a wrapper around `ingest/3` and `to_multi/1`.\n\nExample:\n\n    %Ecto.Multi{} = multi =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todos.Todo, check: &my_check_function/1)\n      |> Phoenix.Sync.Writer.allow(MyApp.Options.Option, check: &my_check_function/1)\n      |> Phoenix.Sync.Writer.to_multi(changes, format: Phoenix.Sync.Writer.Format.TanstackDB)","ref":"Phoenix.Sync.Writer.html#to_multi/3"},{"type":"function","title":"Phoenix.Sync.Writer.transact/4","doc":"Apply operations from a mutation directly via a transaction.\n\n`operation_fun` is a 1-arity function that receives each of the\n`%Phoenix.Sync.Writer.Operation{}` structs within the mutation data and\nshould apply them appropriately. It should return `:ok` or `{:ok, result}` if\nsuccessful or `{:error, reason}` if the operation is invalid or failed to\napply. If any operation returns `{:error, _}` or raises then the entire\ntransaction is aborted.\n\nThis function will return `{:error, reason}` if the transaction data fails to parse.\n\n    {:ok, txid} =\n      Phoenix.Sync.Writer.transact(\n        my_encoded_txn,\n        MyApp.Repo,\n        fn\n          %{operation: :insert, relation: [_, \"todos\"], change: change} ->\n            MyApp.Repo.insert(...)\n          %{operation: :update, relation: [_, \"todos\"], data: data, change: change} ->\n            MyApp.Repo.update(Ecto.Changeset.cast(...))\n          %{operation: :delete, relation: [_, \"todos\"], data: data} ->\n            # we don't allow deletes...\n            {:error, \"invalid delete\"}\n        end,\n        format: Phoenix.Sync.Writer.Format.TanstackDB,\n        timeout: 60_000\n      )\n\nAny of the `opts` not used by this module are passed onto the\n`c:Ecto.Repo.transaction/2` call.\n\nThis is equivalent to the below:\n\n    {:ok, txn} =\n      Phoenix.Sync.Writer.parse_transaction(\n        my_encoded_txn,\n        format: Phoenix.Sync.Writer.Format.TanstackDB\n      )\n\n    {:ok, txid} =\n      MyApp.Repo.transaction(fn ->\n        Enum.each(txn.operations, fn\n          %{operation: :insert, relation: [_, \"todos\"], change: change} ->\n            # insert a Todo\n          %{operation: :update, relation: [_, \"todos\"], data: data, change: change} ->\n            # update a Todo\n          %{operation: :delete, relation: [_, \"todos\"], data: data} ->\n            # we don't allow deletes...\n            raise \"invalid delete\"\n        end)\n        Phoenix.Sync.Writer.txid!(MyApp.Repo)\n      end, timeout: 60_000)","ref":"Phoenix.Sync.Writer.html#transact/4"},{"type":"type","title":"Phoenix.Sync.Writer.transact_opts/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:transact_opts/0"},{"type":"function","title":"Phoenix.Sync.Writer.transaction/3","doc":"Runs the mutation inside a transaction.\n\nSince the mutation operation is expressed as an `Ecto.Multi` operation, see\nthe [`Ecto.Repo` docs](https://hexdocs.pm/ecto/Ecto.Repo.html#c:transaction/2-use-with-ecto-multi)\nfor the result if any of your mutations returns an error.\n\n    Phoenix.Sync.Writer.new()\n    |> Phoenix.Sync.Writer.allow(MyApp.Todos.Todo)\n    |> Phoenix.Sync.Writer.allow(MyApp.Options.Option)\n    |> Phoenix.Sync.Writer.ingest(\n      changes,\n      format: Phoenix.Sync.Writer.Format.TanstackDB\n    )\n    |> Phoenix.Sync.Writer.transaction(MyApp.Repo)\n    |> case do\n      {:ok, txid, _changes} ->\n        # return the txid to the client\n        Plug.Conn.send_resp(conn, 200, Jason.encode!(%{txid: txid}))\n      {:error, _failed_operation, failed_value, _changes_so_far} ->\n        # extract the error message from the changeset returned as `failed_value`\n        error =\n          Ecto.Changeset.traverse_errors(failed_value, fn {msg, opts} ->\n            Regex.replace(~r\"%{(w+)}\", msg, fn _, key ->\n              opts |> Keyword.get(String.to_existing_atom(key), key) |> to_string()\n            end)\n          end)\n        Plug.Conn.send_resp(conn, 400, Jason.encode!(error))\n      end\n\nAlso supports normal fun/0 or fun/1 style transactions much like\n`c:Ecto.Repo.transaction/2`, returning the txid of the operation:\n\n    {:ok, txid, todo} =\n      Phoenix.Sync.Writer.transaction(fn ->\n        Repo.insert!(changeset)\n      end, Repo)","ref":"Phoenix.Sync.Writer.html#transaction/3"},{"type":"type","title":"Phoenix.Sync.Writer.txid/0","doc":"","ref":"Phoenix.Sync.Writer.html#t:txid/0"},{"type":"function","title":"Phoenix.Sync.Writer.txid/1","doc":"Extract the transaction id from changes or from a `Ecto.Repo` within a\ntransaction.\n\nThis allows you to use a standard `c:Ecto.Repo.transaction/2` call to apply\nmutations defined using `apply/2` and extract the transaction id afterwards.\n\nExample\n\n    {:ok, changes} =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(MyApp.Todos.Todo)\n      |> Phoenix.Sync.Writer.allow(MyApp.Options.Option)\n      |> Phoenix.Sync.Writer.to_multi(changes, format: Phoenix.Sync.Writer.Format.TanstackDB)\n      |> MyApp.Repo.transaction()\n\n    {:ok, txid} = Phoenix.Sync.Writer.txid(changes)\n\nIt also allows you to get a transaction id from any active transaction:\n\n    MyApp.Repo.transaction(fn ->\n      {:ok, txid} = Phoenix.Sync.Writer.txid(MyApp.Repo)\n    end)\n\nAttempting to run `txid/1` on a repo outside a transaction will return an\nerror.","ref":"Phoenix.Sync.Writer.html#txid/1"},{"type":"function","title":"Phoenix.Sync.Writer.txid!/1","doc":"Returns the a transaction id or raises on an error.\n\nSee `txid/1`.","ref":"Phoenix.Sync.Writer.html#txid!/1"},{"type":"module","title":"Phoenix.Sync.Writer.Context","doc":"Provides context within callback functions.","ref":"Phoenix.Sync.Writer.Context.html"},{"type":"module","title":"Fields - Phoenix.Sync.Writer.Context","doc":"* `index` - the 0-indexed position of the current operation within the transaction\n  * `changes` - the current `Ecto.Multi` changes so far\n  * `operation` - the current `Phoenix.Sync.Writer.Context.Operation`\n  * `callback` - the name of the current callback, `:pre_apply` or `:post_apply`\n  * `schema` - the `Ecto.Schema` module associated with the current operation","ref":"Phoenix.Sync.Writer.Context.html#module-fields"},{"type":"behaviour","title":"Phoenix.Sync.Writer.Format","doc":"Defines a behaviour that applications can implement in order to handle custom\ndata formats within a `Phoenix.Sync.Writer`\n\nThe exact format of the change messages coming from the client is\nunimportant, however `Phoenix.Sync.Writer` requires certain essential\ninformation that needs to be included.\n\n- `operation` - one of `insert`, `update` or `delete`\n- `relation` - the table name that the operation applies to.\n- `data` - the original data before the mutation was applied. **Required** for\n  `update` or `delete` operations\n- `changes` - the changes to apply. **Required** for `insert` and `update` operations\n\nHowever you choose to encode this information on the client, you simply need\nto set the `format` of your write configuration using to a module that\nimplements this behaviour.\n\n#### Implementation guide\n\nOnce you have parsed the data coming from the client, over HTTP or even raw\nTCP, use `Phoenix.Sync.Writer.Operation.new/4` (or\n`Phoenix.Sync.Writer.Transaction.operation/4`) to validate the values and\ncreate a `%Phoenix.Sync.Writer.Operation{}` struct.\n\n`Phoenix.Sync.Writer.Transaction.parse_operations/2` is a helper function for\nerror handling when mapping a list of encoded operation data using\n`Phoenix.Sync.Writer.Transaction.operation/4`.","ref":"Phoenix.Sync.Writer.Format.html"},{"type":"behaviour","title":"Example: - Phoenix.Sync.Writer.Format","doc":"We use Protocol buffers to encode the incoming information using `:protox`\nand implement this module's `c:parse_transaction/1` callback using the `Protox` decode functions:\n\n    defmodule MyApp.Protocol do\n      use Protox,\n        namespace: __MODULE__,\n        schema: ~S[\n          syntax: \"proto2\";\n          message Operation {\n            enum Op {\n              INSERT = 0;\n              UPDATE = 1;\n              DELETE = 2;\n            }\n            required Op op = 1;\n            string table = 2;\n            map  original = 3;\n            map  modified = 4;\n          }\n          message Transaction {\n            repeated Operation operations = 1;\n          }\n        ]\n\n      alias Phoenix.Sync.Writer\n\n      @behaviour Phoenix.Sync.Writer.Format\n\n      @impl Phoenix.Sync.Writer.Format\n      def parse_transaction(proto) when is_binary(proto) do\n        with {:ok, %{operations: ops}} <- MyApp.Protocol.Transaction.decode(proto),\n             {:ok, operations} <- Writer.Transaction.parse_operations(ops, &convert_operation/1) do\n          {:ok, %Writer.Transaction{operations: operations}\n        end\n      end\n\n      defp convert_operation(%MyApp.Protocol.Operation{} = operation) do\n        Writer.Transaction.operation(\n          operation.op,\n          operation.table,\n          operation.original,\n          operation.modified\n        )\n      end\n    end\n\nWe use our protocol module as the `format` when we `apply/4` our transaction\ndata and we can just pass our serialized protobuf message as the mutation data:\n\n    {:ok, txid, _changes} =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.apply(protobuf_data, MyApp.Repo, format: MyApp.Protocol)","ref":"Phoenix.Sync.Writer.Format.html#module-example"},{"type":"callback","title":"Phoenix.Sync.Writer.Format.parse_transaction/1","doc":"Translate some data format into a `Phoenix.Sync.Writer.Transaction` with a\nlist of operations to apply.","ref":"Phoenix.Sync.Writer.Format.html#c:parse_transaction/1"},{"type":"type","title":"Phoenix.Sync.Writer.Format.parse_transaction_result/0","doc":"","ref":"Phoenix.Sync.Writer.Format.html#t:parse_transaction_result/0"},{"type":"type","title":"Phoenix.Sync.Writer.Format.parser_fun/0","doc":"","ref":"Phoenix.Sync.Writer.Format.html#t:parser_fun/0"},{"type":"type","title":"Phoenix.Sync.Writer.Format.t/0","doc":"","ref":"Phoenix.Sync.Writer.Format.html#t:t/0"},{"type":"type","title":"Phoenix.Sync.Writer.Format.transaction_data/0","doc":"Raw data from a client that will be parsed to a `Phoenix.Sync.Writer.Format.Transaction` by the Writer's format parser","ref":"Phoenix.Sync.Writer.Format.html#t:transaction_data/0"},{"type":"module","title":"Phoenix.Sync.Writer.Format.TanstackDB","doc":"Implements the `Phoenix.Sync.Writer.Format` behaviour for the data format used by\n[TanStack/db](https://github.com/TanStack/db).","ref":"Phoenix.Sync.Writer.Format.TanstackDB.html"},{"type":"function","title":"Phoenix.Sync.Writer.Format.TanstackDB.parse_operation/1","doc":"","ref":"Phoenix.Sync.Writer.Format.TanstackDB.html#parse_operation/1"},{"type":"module","title":"Phoenix.Sync.Writer.Operation","doc":"Represents a mutation operation received from a client.\n\nTo handle custom formats, translate incoming changes into a [`Operation`](`Phoenix.Sync.Writer.Operation`)\nstruct using `new/4`.","ref":"Phoenix.Sync.Writer.Operation.html"},{"type":"function","title":"Phoenix.Sync.Writer.Operation.new/4","doc":"Takes data from a mutation and validates it before returning a struct.","ref":"Phoenix.Sync.Writer.Operation.html#new/4"},{"type":"function","title":"Parameters - Phoenix.Sync.Writer.Operation.new/4","doc":"- `operation` one of `\"insert\"`,`\"INSERT\"`,`:insert`,`:INSERT`,`\"update\"`,`\"UPDATE\"`,`:update`,`:UPDATE`,`\"delete\"`,`\"DELETE\"`,`:delete`,`:DELETE`\n- `table` the client table name for the write. Can either be a plain string\n  name `\"table\"` or a list with `[\"schema\", \"table\"]`.\n- `data` the original values (see [Updates vs Inserts vs Deletes](#new/4-updates-vs-inserts-vs-deletes))\n- `changes` any updates to apply (see [Updates vs Inserts vs Deletes](#new/4-updates-vs-inserts-vs-deletes))","ref":"Phoenix.Sync.Writer.Operation.html#new/4-parameters"},{"type":"function","title":"Updates vs Inserts vs Deletes - Phoenix.Sync.Writer.Operation.new/4","doc":"The `Phoenix.Sync.Writer.Operation` struct has two value fields, `data` and `changes`.\n\n`data` represents what's already in the database, and `changes` what's\ngoing to be written over the top of this.\n\nFor `insert` operations, `data` is ignored so the new values for the\ninserted row should be in `changes`.\n\nFor `deletes`, `changes` is ignored and `data` should contain the row\nspecification to delete. This needn't be the full row, but must contain\nvalues for all the **primary keys** for the table.\n\nFor `updates`, `data` should contain the original row values and `changes`\nthe changed fields.\n\nThese fields map to the arguments `Ecto.Changeset.change/2` and\n`Ecto.Changeset.cast/4` functions, `data` is used to populate the first\nargument of these functions and `changes` the second.","ref":"Phoenix.Sync.Writer.Operation.html#new/4-updates-vs-inserts-vs-deletes"},{"type":"function","title":"Phoenix.Sync.Writer.Operation.new!/4","doc":"","ref":"Phoenix.Sync.Writer.Operation.html#new!/4"},{"type":"type","title":"Phoenix.Sync.Writer.Operation.new_result/0","doc":"","ref":"Phoenix.Sync.Writer.Operation.html#t:new_result/0"},{"type":"type","title":"Phoenix.Sync.Writer.Operation.t/0","doc":"","ref":"Phoenix.Sync.Writer.Operation.html#t:t/0"},{"type":"module","title":"Phoenix.Sync.Writer.Transaction","doc":"Represents a transaction containing a list of `Phoenix.Sync.Writer.Operation`s\nthat should be applied atomically.\n\n```elixir\n{:ok, operations} <- Transaction.parse_operations(operations, &parse_operation/1)\n\n%Transaction{} = Transaction.new(operations)\n```","ref":"Phoenix.Sync.Writer.Transaction.html"},{"type":"function","title":"Phoenix.Sync.Writer.Transaction.empty/0","doc":"Return a new, empty, `Transaction` struct.","ref":"Phoenix.Sync.Writer.Transaction.html#empty/0"},{"type":"type","title":"Phoenix.Sync.Writer.Transaction.id/0","doc":"","ref":"Phoenix.Sync.Writer.Transaction.html#t:id/0"},{"type":"function","title":"Phoenix.Sync.Writer.Transaction.new/1","doc":"","ref":"Phoenix.Sync.Writer.Transaction.html#new/1"},{"type":"function","title":"Phoenix.Sync.Writer.Transaction.operation/4","doc":"","ref":"Phoenix.Sync.Writer.Transaction.html#operation/4"},{"type":"function","title":"Phoenix.Sync.Writer.Transaction.operation!/4","doc":"","ref":"Phoenix.Sync.Writer.Transaction.html#operation!/4"},{"type":"function","title":"Phoenix.Sync.Writer.Transaction.parse_operations/2","doc":"Helper function to parse a list of encoded Operations.","ref":"Phoenix.Sync.Writer.Transaction.html#parse_operations/2"},{"type":"type","title":"Phoenix.Sync.Writer.Transaction.t/0","doc":"","ref":"Phoenix.Sync.Writer.Transaction.html#t:t/0"},{"type":"extras","title":"Phoenix.Sync","doc":"# Phoenix.Sync\n\nReal-time sync for Postgres-backed [Phoenix](https://www.phoenixframework.org/) applications.\n\n \n   \n     \n       \n     \n   \n \n\n[![Hex.pm](https://img.shields.io/hexpm/v/phoenix_sync.svg)](https://hex.pm/packages/phoenix_sync)\n[![Documentation](https://img.shields.io/badge/docs-hexdocs-green)](https://hexdocs.pm/phoenix_sync)\n[![License](https://img.shields.io/badge/license-Apache_2.0-green)](./LICENSE)\n[![Status](https://img.shields.io/badge/status-beta-orange)](https://github.com/electric-sql/phoenix_sync)\n[![Discord](https://img.shields.io/discord/933657521581858818?color=5969EA&label=discord)](https://discord.electric-sql.com)\n\nDocumentation is available at [hexdocs.pm/phoenix_sync](https://hexdocs.pm/phoenix_sync).","ref":"readme.html"},{"type":"extras","title":"Build real-time apps on sync - Phoenix.Sync","doc":"Phoenix.Sync is a library that adds real-time sync to Postgres-backed [Phoenix](https://www.phoenixframework.org/) applications. Use it to sync data into both LiveView and front-end web and mobile applications.\n\n- integrates with `Plug` and `Phoenix.{Controller, LiveView, Router, Stream}`\n- uses [ElectricSQL](https://electric-sql.com) for core sync, fan-out and data delivery\n- maps `Ecto.Query`s to [Shapes](https://electric-sql.com/docs/guides/shapes) for partial replication\n\nThere are four key APIs for [read-path sync](#read-path-sync) out of Postgres:\n\n- `Phoenix.Sync.Client.stream/2` for low level usage in Elixir\n- `Phoenix.Sync.LiveView.sync_stream/4` to sync into a LiveView\n- `Phoenix.Sync.Router.sync/2` macro to expose a shape in your Router\n- `Phoenix.Sync.Controller.sync_render/3` to return shapes from a Controller\n\nAnd a `Phoenix.Sync.Writer` module for handling [write-path sync](#write-path-sync) back into Postgres.","ref":"readme.html#build-real-time-apps-on-sync"},{"type":"extras","title":"Read-path sync - Phoenix.Sync","doc":"","ref":"readme.html#read-path-sync"},{"type":"extras","title":"Low level usage in Elixir - Phoenix.Sync","doc":"Use `Phoenix.Sync.Client.stream/2` to convert an `Ecto.Query` into an Elixir `Stream`:\n\n```elixir\nstream = Phoenix.Sync.Client.stream(Todos.Todo)\n\nstream =\n  Ecto.Query.from(t in Todos.Todo, where: t.completed == false)\n  |> Phoenix.Sync.Client.stream()\n```","ref":"readme.html#low-level-usage-in-elixir"},{"type":"extras","title":"Sync into a LiveView stream - Phoenix.Sync","doc":"Swap out `Phoenix.LiveView.stream/3` for `Phoenix.Sync.LiveView.sync_stream/4` to automatically keep a LiveView up-to-date with the state of your Postgres database:\n\n```elixir\ndefmodule MyWeb.MyLive do\n  use Phoenix.LiveView\n  import Phoenix.Sync.LiveView\n\n  def mount(_params, _session, socket) do\n    {:ok, sync_stream(socket, :todos, Todos.Todo)}\n  end\n\n  def handle_info({:sync, event}, socket) do\n    {:noreply, sync_stream_update(socket, event)}\n  end\nend\n```\n\nLiveView takes care of automatically keeping the front-end up-to-date with the assigned stream. What Phoenix.Sync does is automatically keep the _stream_ up-to-date with the state of the database.\n\nThis means you can build fully end-to-end real-time multi-user applications without writing Javascript _and_ without worrying about message delivery, reconnections, cache invalidation or polling the database for changes.","ref":"readme.html#sync-into-a-liveview-stream"},{"type":"extras","title":"Sync shapes through your Router - Phoenix.Sync","doc":"Use the `Phoenix.Sync.Router.sync/2` macro to expose statically (compile-time) defined shapes in your Router:\n\n```elixir\ndefmodule MyWeb.Router do\n  use Phoenix.Router\n  import Phoenix.Sync.Router\n\n  pipeline :sync do\n    plug :my_auth\n  end\n\n  scope \"/shapes\" do\n    pipe_through :sync\n\n    sync \"/todos\", Todos.Todo\n  end\nend\n```\n\nBecause the shapes are exposed through your Router, the client connects through your existing Plug middleware. This allows you to do real-time sync straight out of Postgres _without_ having to translate your auth logic into complex/fragile database rules.","ref":"readme.html#sync-shapes-through-your-router"},{"type":"extras","title":"Sync dynamic shapes from a Controller - Phoenix.Sync","doc":"Sync shapes from any standard Controller using the `Phoenix.Sync.Controller.sync_render/3` view function:\n\n```elixir\ndefmodule Phoenix.Sync.LiveViewTest.TodoController do\n  use Phoenix.Controller\n  import Phoenix.Sync.Controller\n  import Ecto.Query, only: [from: 2]\n\n  def show(conn, %{\"done\" => done} = params) do\n    sync_render(conn, params, from(t in Todos.Todo, where: t.done == ^done))\n  end\n\n  def show_mine(%{assigns: %{current_user: user_id}} = conn, params) do\n    sync_render(conn, params, from(t in Todos.Todo, where: t.owner_id == ^user_id))\n  end\nend\n```\n\nThis allows you to define and personalise the shape definition at runtime using the session and request.","ref":"readme.html#sync-dynamic-shapes-from-a-controller"},{"type":"extras","title":"Consume shapes in the frontend - Phoenix.Sync","doc":"You can sync _into_ any client in any language that [speaks HTTP and JSON](https://electric-sql.com/docs/api/http).\n\nFor example, using the Electric [TypeScript client](https://electric-sql.com/docs/api/clients/typescript):\n\n```typescript\nimport { Shape, ShapeStream } from \"@electric-sql/client\";\n\nconst stream = new ShapeStream({\n  url: `/shapes/todos`,\n});\nconst shape = new Shape(stream);\n\n// The callback runs every time the data changes.\nshape.subscribe((data) => console.log(data));\n```\n\nOr binding a shape to a component using the [React bindings](https://electric-sql.com/docs/integrations/react):\n\n```tsx\nimport { useShape } from \"@electric-sql/react\";\n\nconst MyComponent = () => {\n  const { data } = useShape({\n    url: `shapes/todos`,\n  });\n\n  return  ;\n};\n```\n\nSee the Electric [demos](https://electric-sql.com/demos) and [documentation](https://electric-sql.com/docs) for more client-side usage examples.","ref":"readme.html#consume-shapes-in-the-frontend"},{"type":"extras","title":"Write-path sync - Phoenix.Sync","doc":"The `Phoenix.Sync.Writer` module allows you to ingest batches of writes from the client.\n\nThe idea is that the front-end can batch up [local optimistic writes](https://electric-sql.com/docs/guides/writes). For example using a library like [@TanStack/db](https://github.com/TanStack/db) or by [monitoring changes to a local embedded database](https://electric-sql.com/docs/guides/writes#through-the-db).\n\nThese changes can be POSTed to a `Phoenix.Controller`, which then constructs a `Phoenix.Sync.Writer` instance. The writer instance authorizes and validates the writes before applying them to the database. Under the hood this uses `Ecto.Multi`, to ensure that transactions (batches of writes) are applied atomically.\n\nFor example, the controller below handles local writes made to a project management app. It constructs a writer instance and pipes it through a series of `Phoenix.Sync.Writer.allow/3` calls. These register functions against `Ecto.Schema`s (in this case `Projects.Project` and `Projects.Issue`):\n\n```elixir\ndefmodule MutationController do\n  use Phoenix.Controller, formats: [:json]\n\n  alias Phoenix.Sync.Writer\n  alias Phoenix.Sync.Writer.Format\n\n  def mutate(conn, %{\"transaction\" => transaction} = _params) do\n    user_id = conn.assigns.user_id\n\n    {:ok, txid, _changes} =\n      Phoenix.Sync.Writer.new()\n      |> Phoenix.Sync.Writer.allow(\n        Projects.Project,\n        check: reject_invalid_params/2,\n        load: &Projects.load_for_user(&1, user_id),\n        validate: &Projects.Project.changeset/2\n      )\n      |> Phoenix.Sync.Writer.allow(\n        Projects.Issue,\n        # Use the sensible defaults:\n        # validate: Projects.Issue.changeset/2\n        # etc.\n      )\n      |> Phoenix.Sync.Writer.apply(transaction, Repo, format: Format.TanstackDB)\n\n    render(conn, :mutations, txid: txid)\n  end\nend\n```\n\nThis facilitates incrementally adding bi-directional sync support to a Phoenix application, re-using your existing auth and schema/validation logic.\n\nSee the `Phoenix.Sync.Writer` module docs for more information.","ref":"readme.html#write-path-sync"},{"type":"extras","title":"Installation and configuration - Phoenix.Sync","doc":"`Phoenix.Sync` can be used in two modes:\n\n1. `:embedded` where Electric is included as an application dependency and Phoenix.Sync consumes data internally using Elixir APIs\n2. `:http` where Electric does _not_ need to be included as an application dependency and Phoenix.Sync consumes data from an external Electric service using it's [HTTP API](https://electric-sql.com/docs/api/http)","ref":"readme.html#installation-and-configuration"},{"type":"extras","title":"Embedded mode - Phoenix.Sync","doc":"In `:embedded` mode, Electric must be included an application dependency but does not expose an HTTP API (internally or externally). Messages are streamed internally between Electric and Phoenix.Sync using Elixir function APIs. The only HTTP API for sync is that exposed via your Phoenix Router using the `sync/2` macro and `sync_render/3` function.\n\nExample config:\n\n```elixir\n# mix.exs\ndefp deps do\n  [\n    {:electric, \"~> 1.0\"},\n    {:phoenix_sync, \"~> 0.4\"}\n  ]\nend\n\n# config/config.exs\nconfig :phoenix_sync,\n  env: config_env(),\n  mode: :embedded,\n  repo: MyApp.Repo\n\n# application.ex\nchildren = [\n  MyApp.Repo,\n  # ...\n  {MyApp.Endpoint, phoenix_sync: Phoenix.Sync.plug_opts()}\n]\n```","ref":"readme.html#embedded-mode"},{"type":"extras","title":"HTTP - Phoenix.Sync","doc":"In `:http` mode, Electric does not need to be included as an application dependency. Instead, Phoenix.Sync consumes data from an external Electric service over HTTP.\n\n```elixir\n# mix.exs\ndefp deps do\n  [\n    {:phoenix_sync, \"~> 0.4\"}\n  ]\nend\n\n# config/config.exs\nconfig :phoenix_sync,\n  env: config_env(),\n  mode: :http,\n  url: \"https://api.electric-sql.cloud\",\n  credentials: [\n    secret: \"...\",    # required\n    source_id: \"...\"  # optional, required for Electric Cloud\n  ]\n\n# application.ex\nchildren = [\n  MyApp.Repo,\n  # ...\n  {MyApp.Endpoint, phoenix_sync: Phoenix.Sync.plug_opts()}\n]\n```","ref":"readme.html#http"},{"type":"extras","title":"Local HTTP services - Phoenix.Sync","doc":"It is also possible to include Electric as an application dependency and configure it to expose a local HTTP API that's consumed by Phoenix.Sync running in `:http` mode:\n\n```elixir\n# mix.exs\ndefp deps do\n  [\n    {:electric, \"~> 1.0\"},\n    {:phoenix_sync, \"~> 0.4\"}\n  ]\nend\n\n# config/config.exs\nconfig :phoenix_sync,\n  env: config_env(),\n  mode: :http,\n  http: [\n    port: 3000,\n  ],\n  repo: MyApp.Repo,\n  url: \"http://localhost:3000\"\n\n# application.ex\nchildren = [\n  MyApp.Repo,\n  # ...\n  {MyApp.Endpoint, phoenix_sync: Phoenix.Sync.plug_opts()}\n]\n```\n\nThis is less efficient than running in `:embedded` mode but may be useful for testing or when needing to run an HTTP proxy in front of Electric as part of your development stack.","ref":"readme.html#local-http-services"},{"type":"extras","title":"Different modes for different envs - Phoenix.Sync","doc":"Apps using `:http` mode in certain environments can exclude `:electric` as a dependency for that environment. The following example shows how to configure:\n\n- `:embedded` mode in `:dev`\n- `:http` mode with a local Electric service in `:test`\n- `:http` mode with an external Electric service in `:prod`\n\nWith Electric only included and compiled as a dependency in `:dev` and `:test`.\n\n```elixir\n# mix.exs\ndefp deps do\n  [\n    {:electric, \"~> 1.0\", only: [:dev, :test]},\n    {:phoenix_sync, \"~> 0.4\"}\n  ]\nend\n\n# config/dev.exs\nconfig :phoenix_sync,\n  env: config_env(),\n  mode: :embedded,\n  repo: MyApp.Repo\n\n# config/test.esx\nconfig :phoenix_sync,\n  env: config_env(),\n  mode: :http,\n  http: [\n    port: 3000,\n  ],\n  repo: MyApp.Repo,\n  url: \"http://localhost:3000\"\n\n# config/prod.exs\nconfig :phoenix_sync,\n  mode: :http,\n  url: \"https://api.electric-sql.cloud\",\n  credentials: [\n    secret: \"...\",    # required\n    source_id: \"...\"  # optional, required for Electric Cloud\n  ]\n\n# application.ex\nchildren = [\n  MyApp.Repo,\n  # ...\n  {MyApp.Endpoint, phoenix_sync: Phoenix.Sync.plug_opts()}\n]\n```","ref":"readme.html#different-modes-for-different-envs"},{"type":"extras","title":"Notes - Phoenix.Sync","doc":"","ref":"readme.html#notes"},{"type":"extras","title":"ElectricSQL - Phoenix.Sync","doc":"ElectricSQL is a [real-time sync engine for Postgres](https://electric-sql.com).\n\nPhoenix.Sync uses Electric to handle the core concerns of partial replication, fan out and data delivery.","ref":"readme.html#electricsql"},{"type":"extras","title":"Partial replication - Phoenix.Sync","doc":"Electric defines partial replication using [Shapes](https://electric-sql.com/docs/guides/shapes).","ref":"readme.html#partial-replication"},{"type":"extras","title":"Shape definitions - Phoenix.Sync","doc":"Phoenix.Sync allows shapes to be defined in two ways:\n\n1. using an `Ecto.Query` (or `Ecto.Schema` module)\n2. using a keyword list","ref":"readme.html#shape-definitions"},{"type":"extras","title":"Using an `Ecto.Query` - Phoenix.Sync","doc":"Phoenix.Sync maps Ecto queries to shape definitions. This allows you to control what data syncs where using Ecto.Schema and Ecto.Query.\n\nFor example, using a query:\n\n```elixir\nsync_render(conn, params, from(t in Todos.Todo, where: t.completed == false))\n```\n\nOr using a schema (equivalent to `from(t in Todos.Todo)`):\n\n```elixir\nsync_render(conn, params, Todos.Todo)\n```\n\nQuery support is currently limited to `where` conditions. Support for more complex queries using `join`, `order_by`, `limit` and preloaded association graphs is planned and will be added in Q2 2025.\n\nThe static shapes defined using the `sync/2` or `sync/3` router macros do not accept `Ecto.Query` structs as a shape definition. This is to avoid excessive recompilation caused by your router having a compile-time dependency on your `Ecto` schemas.\n\nIf you want to add a where-clause filter to a static shape in your router, you can add an explicit [`where` clause](https://electric-sql.com/docs/guides/shapes#where-clause) alongside your `Ecto.Schema` module:\n\n```elixir\n\nsync \"/incomplete-todos\", Todos.Todo, where: \"completed = false\"\n```\n\nYou can also include `replica` (see below) in your static shape definitions:\n\n```elixir\n\nsync \"/incomplete-todos\", Todos.Todo, where: \"completed = false\", replica: :full\n```\n\nFor anything else more dynamic, or to use Ecto queries, you should switch from using the `sync` macros in your router to using `sync_render/3` in a controller.","ref":"readme.html#using-an-ecto-query"},{"type":"extras","title":"Using a keyword list - Phoenix.Sync","doc":"At minimum a shape requires a `table`. You can think of shapes defined with\njust a table name as the sync-equivalent of `SELECT * FROM table`.\n\nThe available options are:\n\n- `table` (required). The Postgres table name. Be aware of casing and [Postgres's handling of unquoted upper-case names](https://wiki.postgresql.org/wiki/Don%27t_Do_This#Don.27t_use_upper_case_table_or_column_names).\n- `namespace` (optional). The Postgres namespace that the table belongs to. Defaults to `public`.\n- `where` (optional). Filter to apply to the synced data in SQL format, e.g. `where: \"amount < 1.23 AND colour in ('red', 'green')\"`.\n- `columns` (optional). The columns to include in the synced data. By default Electric will include all columns in the table. The column list **must** include all primary keys. E.g. `columns: [\"id\", \"title\", \"amount\"]`.\n- `replica` (optional). By default Electric will only send primary keys + changed columns on updates. Set `replica: :full` to receive the full row, not just the changed columns.\n\nSee the [Electric Shapes guide](https://electric-sql.com/docs/guides/shapes) for more information.","ref":"readme.html#using-a-keyword-list"},{"type":"extras","title":"LICENSE","doc":"Apache License\n                           Version 2.0, January 2004\n                        http://www.apache.org/licenses/\n\n   TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION\n\n   1. Definitions.\n\n      \"License\" shall mean the terms and conditions for use, reproduction,\n      and distribution as defined by Sections 1 through 9 of this document.\n\n      \"Licensor\" shall mean the copyright owner or entity authorized by\n      the copyright owner that is granting the License.\n\n      \"Legal Entity\" shall mean the union of the acting entity and all\n      other entities that control, are controlled by, or are under common\n      control with that entity. For the purposes of this definition,\n      \"control\" means (i) the power, direct or indirect, to cause the\n      direction or management of such entity, whether by contract or\n      otherwise, or (ii) ownership of fifty percent (50%) or more of the\n      outstanding shares, or (iii) beneficial ownership of such entity.\n\n      \"You\" (or \"Your\") shall mean an individual or Legal Entity\n      exercising permissions granted by this License.\n\n      \"Source\" form shall mean the preferred form for making modifications,\n      including but not limited to software source code, documentation\n      source, and configuration files.\n\n      \"Object\" form shall mean any form resulting from mechanical\n      transformation or translation of a Source form, including but\n      not limited to compiled object code, generated documentation,\n      and conversions to other media types.\n\n      \"Work\" shall mean the work of authorship, whether in Source or\n      Object form, made available under the License, as indicated by a\n      copyright notice that is included in or attached to the work\n      (an example is provided in the Appendix below).\n\n      \"Derivative Works\" shall mean any work, whether in Source or Object\n      form, that is based on (or derived from) the Work and for which the\n      editorial revisions, annotations, elaborations, or other modifications\n      represent, as a whole, an original work of authorship. For the purposes\n      of this License, Derivative Works shall not include works that remain\n      separable from, or merely link (or bind by name) to the interfaces of,\n      the Work and Derivative Works thereof.\n\n      \"Contribution\" shall mean any work of authorship, including\n      the original version of the Work and any modifications or additions\n      to that Work or Derivative Works thereof, that is intentionally\n      submitted to Licensor for inclusion in the Work by the copyright owner\n      or by an individual or Legal Entity authorized to submit on behalf of\n      the copyright owner. For the purposes of this definition, \"submitted\"\n      means any form of electronic, verbal, or written communication sent\n      to the Licensor or its representatives, including but not limited to\n      communication on electronic mailing lists, source code control systems,\n      and issue tracking systems that are managed by, or on behalf of, the\n      Licensor for the purpose of discussing and improving the Work, but\n      excluding communication that is conspicuously marked or otherwise\n      designated in writing by the copyright owner as \"Not a Contribution.\"\n\n      \"Contributor\" shall mean Licensor and any individual or Legal Entity\n      on behalf of whom a Contribution has been received by Licensor and\n      subsequently incorporated within the Work.\n\n   2. Grant of Copyright License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      copyright license to reproduce, prepare Derivative Works of,\n      publicly display, publicly perform, sublicense, and distribute the\n      Work and such Derivative Works in Source or Object form.\n\n   3. Grant of Patent License. Subject to the terms and conditions of\n      this License, each Contributor hereby grants to You a perpetual,\n      worldwide, non-exclusive, no-charge, royalty-free, irrevocable\n      (except as stated in this section) patent license to make, have made,\n      use, offer to sell, sell, import, and otherwise transfer the Work,\n      where such license applies only to those patent claims licensable\n      by such Contributor that are necessarily infringed by their\n      Contribution(s) alone or by combination of their Contribution(s)\n      with the Work to which such Contribution(s) was submitted. If You\n      institute patent litigation against any entity (including a\n      cross-claim or counterclaim in a lawsuit) alleging that the Work\n      or a Contribution incorporated within the Work constitutes direct\n      or contributory patent infringement, then any patent licenses\n      granted to You under this License for that Work shall terminate\n      as of the date such litigation is filed.\n\n   4. Redistribution. You may reproduce and distribute copies of the\n      Work or Derivative Works thereof in any medium, with or without\n      modifications, and in Source or Object form, provided that You\n      meet the following conditions:\n\n      (a) You must give any other recipients of the Work or\n          Derivative Works a copy of this License; and\n\n      (b) You must cause any modified files to carry prominent notices\n          stating that You changed the files; and\n\n      (c) You must retain, in the Source form of any Derivative Works\n          that You distribute, all copyright, patent, trademark, and\n          attribution notices from the Source form of the Work,\n          excluding those notices that do not pertain to any part of\n          the Derivative Works; and\n\n      (d) If the Work includes a \"NOTICE\" text file as part of its\n          distribution, then any Derivative Works that You distribute must\n          include a readable copy of the attribution notices contained\n          within such NOTICE file, excluding those notices that do not\n          pertain to any part of the Derivative Works, in at least one\n          of the following places: within a NOTICE text file distributed\n          as part of the Derivative Works; within the Source form or\n          documentation, if provided along with the Derivative Works; or,\n          within a display generated by the Derivative Works, if and\n          wherever such third-party notices normally appear. The contents\n          of the NOTICE file are for informational purposes only and\n          do not modify the License. You may add Your own attribution\n          notices within Derivative Works that You distribute, alongside\n          or as an addendum to the NOTICE text from the Work, provided\n          that such additional attribution notices cannot be construed\n          as modifying the License.\n\n      You may add Your own copyright statement to Your modifications and\n      may provide additional or different license terms and conditions\n      for use, reproduction, or distribution of Your modifications, or\n      for any such Derivative Works as a whole, provided Your use,\n      reproduction, and distribution of the Work otherwise complies with\n      the conditions stated in this License.\n\n   5. Submission of Contributions. Unless You explicitly state otherwise,\n      any Contribution intentionally submitted for inclusion in the Work\n      by You to the Licensor shall be under the terms and conditions of\n      this License, without any additional terms or conditions.\n      Notwithstanding the above, nothing herein shall supersede or modify\n      the terms of any separate license agreement you may have executed\n      with Licensor regarding such Contributions.\n\n   6. Trademarks. This License does not grant permission to use the trade\n      names, trademarks, service marks, or product names of the Licensor,\n      except as required for reasonable and customary use in describing the\n      origin of the Work and reproducing the content of the NOTICE file.\n\n   7. Disclaimer of Warranty. Unless required by applicable law or\n      agreed to in writing, Licensor provides the Work (and each\n      Contributor provides its Contributions) on an \"AS IS\" BASIS,\n      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or\n      implied, including, without limitation, any warranties or conditions\n      of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A\n      PARTICULAR PURPOSE. You are solely responsible for determining the\n      appropriateness of using or redistributing the Work and assume any\n      risks associated with Your exercise of permissions under this License.\n\n   8. Limitation of Liability. In no event and under no legal theory,\n      whether in tort (including negligence), contract, or otherwise,\n      unless required by applicable law (such as deliberate and grossly\n      negligent acts) or agreed to in writing, shall any Contributor be\n      liable to You for damages, including any direct, indirect, special,\n      incidental, or consequential damages of any character arising as a\n      result of this License or out of the use or inability to use the\n      Work (including but not limited to damages for loss of goodwill,\n      work stoppage, computer failure or malfunction, or any and all\n      other commercial damages or losses), even if such Contributor\n      has been advised of the possibility of such damages.\n\n   9. Accepting Warranty or Additional Liability. While redistributing\n      the Work or Derivative Works thereof, You may choose to offer,\n      and charge a fee for, acceptance of support, warranty, indemnity,\n      or other liability obligations and/or rights consistent with this\n      License. However, in accepting such obligations, You may act only\n      on Your own behalf and on Your sole responsibility, not on behalf\n      of any other Contributor, and only if You agree to indemnify,\n      defend, and hold each Contributor harmless for any liability\n      incurred by, or claims asserted against, such Contributor by reason\n      of your accepting any such warranty or additional liability.\n\n   END OF TERMS AND CONDITIONS","ref":"license.html"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.37.3"}}